"""
Tools to draw a chord diagram in python
"""

from collections.abc import Sequence

import matplotlib as mpl
import matplotlib.patches as patches

from matplotlib.colors import ColorConverter, Colormap
from matplotlib.path import Path

import numpy as np
import scipy.sparse as ssp

from .gradient import gradient
from .utilities import _get_normed_line, compute_positions, dist, polar2xy


LW = 0.3


def chord_diagram(mat, names=None, order=None, sort="size", directed=False,
                  colors=None, cmap=None, use_gradient=False, chord_colors=None,
                  alpha=0.7, start_at=0, extent=360, width=0.1, pad=2., gap=0.03,
                  chordwidth=0.7, min_chord_width=0, fontsize=12.8,
                  fontcolor="k", rotate_names=False, ax=None, show=False):
    """
    Plot a chord diagram.

    Draws a representation of many-to-many interactions between elements, given
    by an interaction matrix.
    The elements are represented by arcs proportional to their degree and the
    interactions (or fluxes) are drawn as chords joining two arcs:

    * for undirected chords, the size of the arc is proportional to its
      out-degree (or simply its degree if the matrix is fully symmetrical), i.e.
      the sum of the element's row.
    * for directed chords, the size is proportional to the total-degree, i.e.
      the sum of the element's row and column.

    Parameters
    ----------
    mat : square matrix
        Flux data, ``mat[i, j]`` is the flux from i to j.
    names : list of str, optional (default: no names)
        Names of the nodes that will be displayed (must be ordered as the
        matrix entries).
    order : list, optional (default: order of the matrix entries)
        Order in which the arcs should be placed around the trigonometric
        circle.
    sort : str, optional (default: "size")
        Order in which the chords should be sorted: either None (unsorted),
        "size" (default, drawing largest chords first), or "distance"
        (drawing the chords of the two closest arcs at each end of the current
        arc, then progressing towards the connexions with the farthest arcs in
        both drections as we move towards the center of the current arc).
    directed : bool, optional (default: False)
        Whether the chords should be directed, like edges in a graph, with one
        part of each arc dedicated to outgoing chords and the other to incoming
        ones.
    colors : list, optional (default: from `cmap`)
        List of user defined colors or floats.
    cmap : str or colormap object (default: viridis)
        Colormap that will be used to color the arcs and chords by default.
        See `chord_colors` to use different colors for chords.
    use_gradient : bool, optional (default: False)
        Whether a gradient should be use so that chord extremities have the
        same color as the arc they belong to.
    chord_colors : str, or list of colors, optional (default: None)
        Specify color(s) to fill the chords differently from the arcs.
        When the keyword is not used, chord colors default to the colomap given
        by `colors`.
        Possible values for `chord_colors` are:

        * a single color (do not use an RGB tuple, use hex format instead),
          e.g. "red" or "#ff0000"; all chords will have this color
        * a list of colors, e.g. ``["red", "green", "blue"]``, one per node
          (in this case, RGB tuples are accepted as entries to the list).
          Each chord will get its color from its associated source node, or
          from both nodes if `use_gradient` is True.
    alpha : float in [0, 1], optional (default: 0.7)
        Opacity of the chord diagram.
    start_at : float, optional (default : 0)
        Location, in degrees, where the diagram should start on the unit circle.
        Default is to start at 0 degrees, i.e. (x, y) = (1, 0) or 3 o'clock),
        and move counter-clockwise
    extent : float, optional (default : 360)
        The angular aperture, in degrees, of the diagram.
        Default is to use the whole circle, i.e. 360 degrees, but in some cases
        it can be useful to use only a part of it.
    width : float, optional (default: 0.1)
        Width/thickness of the ideogram arc.
    pad : float, optional (default: 2)
        Distance between two neighboring ideogram arcs. Unit: degree.
    gap : float, optional (default: 0)
        Distance between the arc and the beginning of the cord.
    chordwidth : float, optional (default: 0.7)
        Position of the control points for the chords, controlling their shape.
    min_chord_width : float, optional (default: 0)
        Minimal chord width to replace small entries and zero reciprocals in
        the matrix.
    fontsize : float, optional (default: 12.8)
        Size of the fonts for the names.
    fontcolor : str or list, optional (default: black)
        Color of the fonts for the names.
    rotate_names : (list of) bool(s), optional (default: False)
        Whether to rotate all names (if single boolean) or some of them (if
        list) by 90°.
    ax : matplotlib axis, optional (default: new axis)
        Matplotlib axis where the plot should be drawn.
    show : bool, optional (default: False)
        Whether the plot should be displayed immediately via an automatic call
        to `plt.show()`.
    """
    import matplotlib.pyplot as plt

    if ax is None:
        _, ax = plt.subplots()

    # copy matrix
    is_sparse = ssp.issparse(mat)

    if is_sparse:
        mat = ssp.csr_matrix(mat)
    else:
        mat = np.array(mat, copy=True)

    num_nodes = mat.shape[0]

    # don't use gradient with directed chords
    use_gradient *= not directed

    # set min entry size for small entries and zero reciprocals
    # mat[i, j]:  i -> j
    if is_sparse and min_chord_width:
        nnz = mat.nonzero()

        mat.data[mat.data < min_chord_width] = min_chord_width

        # check zero reciprocals
        for i, j in zip(*nnz):
            if mat[j, i] < min_chord_width:
                mat[j, i] = min_chord_width
    elif min_chord_width:
        nnz = mat > 0

        mat[nnz] = np.maximum(mat[nnz], min_chord_width)

        # check zero reciprocals
        for i, j in zip(*np.where(~nnz)):
            if mat[j, i]:
                mat[i, j] = min_chord_width

    # check name rotations
    if isinstance(rotate_names, Sequence):
        assert len(rotate_names) == num_nodes, \
            "Wrong number of entries in 'rotate_names'."
    else:
        rotate_names = [rotate_names]*num_nodes

    # check order
    if order is not None:
        mat = mat[order][:, order]

        rotate_names = [rotate_names[i] for i in order]

        if names is not None:
            names = [names[i] for i in order]

        if colors is not None:
            colors = [colors[i] for i in order]

    # configure colors
    if colors is None:
        colors = np.linspace(0, 1, num_nodes)

    if isinstance(fontcolor, str):
        fontcolor = [fontcolor]*num_nodes
    else:
        assert len(fontcolor) == num_nodes, \
            "One fontcolor per node is required."

    if cmap is None:
        cmap = mpl.colormaps["viridis"]
    elif not isinstance(cmap, Colormap):
        cmap = mpl.colormaps[cmap]

    if isinstance(colors, (list, tuple, np.ndarray)):
        assert len(colors) == num_nodes, "One color per node is required."

        # check color type
        first_color = colors[0]

        if isinstance(first_color, (int, float, np.integer)):
            colors = cmap(colors)[:, :3]
        else:
            colors = [ColorConverter.to_rgb(c) for c in colors]
    else:
        raise ValueError("`colors` should be a list.")

    if chord_colors is None:
       chord_colors = colors
    else:
        try:
            chord_colors = [ColorConverter.to_rgb(chord_colors)] * num_nodes
        except ValueError:
            assert len(chord_colors) == num_nodes, \
                "If `chord_colors` is a list of colors, it should include " \
                "one color per node (here {} colors).".format(num_nodes)

    # sum over rows
    out_deg = mat.sum(axis=1).A1 if is_sparse else mat.sum(axis=1)
    in_deg = None
    degree = out_deg.copy()

    if directed:
        # also sum over columns
        in_deg = mat.sum(axis=0).A1 if is_sparse else mat.sum(axis=0)
        degree += in_deg

    pos = {}
    pos_dir = {}
    arc = []
    nodePos = []
    rotation = []

    # compute all values and optionally apply sort
    compute_positions(mat, degree, in_deg, out_deg, start_at, is_sparse, sort,
                      directed, extent, pad, arc, rotation, nodePos, pos)

    # plot
    for i in range(num_nodes):
        color = colors[i]

        # plot the arcs
        start_at, end = arc[i]

        ideogram_arc(start=start_at, end=end, radius=1.0, color=color,
                     width=width, alpha=alpha, ax=ax)

        chord_color = chord_colors[i]

        # plot self-chords if directed is False
        if not directed and mat[i, i]:
            start1, end1, _, _ = pos[(i, i)]
            self_chord_arc(start1, end1, radius=1 - width - gap,
                           chordwidth=0.7*chordwidth, color=chord_color,
                           alpha=alpha, ax=ax)

        # plot all other chords
        targets = range(num_nodes) if directed else range(i)

        for j in targets:
            cend = chord_colors[j]

            start1, end1, start2, end2 = pos[(i, j)]

            if mat[i, j] > 0 or (not directed and mat[j, i] > 0):
                chord_arc(
                    start1, end1, start2, end2, radius=1 - width - gap, gap=gap,
                    chordwidth=chordwidth, color=chord_color, cend=cend,
                    alpha=alpha, ax=ax, use_gradient=use_gradient,
                    extent=extent, directed=directed)

    # add names if necessary
    if names is not None:
        assert len(names) == num_nodes, "One name per node is required."

        prop = {
            "fontsize": fontsize,
            "ha": "center",
            "va": "center",
            "rotation_mode": "anchor"
        }

        for i, (pos, name, r) in enumerate(zip(nodePos, names, rotation)):
            rotate = rotate_names[i]
            pp = prop.copy()
            pp["color"] = fontcolor[i]

            if rotate:
                angle  = np.average(arc[i])
                rotate = 90

                if 90 < angle < 180 or 270 < angle:
                    rotate = -90

                if 90 < angle < 270:
                    pp["ha"] = "right"
                else:
                    pp["ha"] = "left"
            elif r:
                pp["va"] = "top"
            else:
                pp["va"] = "bottom"

            ax.text(pos[0], pos[1], name, rotation=pos[2] + rotate, **pp)

    # configure axis
    ax.set_xlim(-1.1, 1.1)
    ax.set_ylim(-1.1, 1.1)

    ax.set_aspect(1)
    ax.axis('off')

    plt.tight_layout()

    if show:
        plt.show()

    return nodePos


# ------------ #
# Subfunctions #
# ------------ #

def initial_path(start, end, radius, width, factor=4/3):
    ''' First 16 vertices and 15 instructions are the same for everyone '''
    if start > end:
        start, end = end, start

    start *= np.pi/180.
    end   *= np.pi/180.

    # optimal distance to the control points
    # https://stackoverflow.com/questions/1734745/
    # how-to-create-circle-with-b%C3%A9zier-curves
    # use 16-vertex curves (4 quadratic Beziers which accounts for worst case
    # scenario of 360 degrees)
    inner = radius*(1-width)
    opt   = factor * np.tan((end-start)/ 16.) * radius
    inter1 = start*(3./4.)+end*(1./4.)
    inter2 = start*(2./4.)+end*(2./4.)
    inter3 = start*(1./4.)+end*(3./4.)

    verts = [
        polar2xy(radius, start),
        polar2xy(radius, start) + polar2xy(opt, start+0.5*np.pi),
        polar2xy(radius, inter1) + polar2xy(opt, inter1-0.5*np.pi),
        polar2xy(radius, inter1),
        polar2xy(radius, inter1),
        polar2xy(radius, inter1) + polar2xy(opt, inter1+0.5*np.pi),
        polar2xy(radius, inter2) + polar2xy(opt, inter2-0.5*np.pi),
        polar2xy(radius, inter2),
        polar2xy(radius, inter2),
        polar2xy(radius, inter2) + polar2xy(opt, inter2+0.5*np.pi),
        polar2xy(radius, inter3) + polar2xy(opt, inter3-0.5*np.pi),
        polar2xy(radius, inter3),
        polar2xy(radius, inter3),
        polar2xy(radius, inter3) + polar2xy(opt, inter3+0.5*np.pi),
        polar2xy(radius, end) + polar2xy(opt, end-0.5*np.pi),
        polar2xy(radius, end)
    ]

    codes = [
        Path.MOVETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.LINETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.LINETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.LINETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
    ]

    return start, end, verts, codes


def ideogram_arc(start, end, radius=1., width=0.2, color="r", alpha=0.7,
                 ax=None):
    '''
    Draw an arc symbolizing a region of the chord diagram.

    Parameters
    ----------
    start : float (degree in 0, 360)
        Starting degree.
    end : float (degree in 0, 360)
        Final degree.
    radius : float, optional (default: 1)
        External radius of the arc.
    width : float, optional (default: 0.2)
        Width of the arc.
    ax : matplotlib axis, optional (default: not plotted)
        Axis on which the arc should be plotted.
    color : valid matplotlib color, optional (default: "r")
        Color of the arc.

    Returns
    -------
    verts, codes : lists
        Vertices and path instructions to draw the shape.
    '''
    start, end, verts, codes = initial_path(start, end, radius, width)

    opt    = 4./3. * np.tan((end-start)/ 16.) * radius
    inner  = radius*(1-width)
    inter1 = start*(3./4.) + end*(1./4.)
    inter2 = start*(2./4.) + end*(2./4.)
    inter3 = start*(1./4.) + end*(3./4.)

    verts += [
        polar2xy(inner, end),
        polar2xy(inner, end) + polar2xy(opt*(1-width), end-0.5*np.pi),
        polar2xy(inner, inter3) + polar2xy(opt*(1-width), inter3+0.5*np.pi),
        polar2xy(inner, inter3),
        polar2xy(inner, inter3),
        polar2xy(inner, inter3) + polar2xy(opt*(1-width), inter3-0.5*np.pi),
        polar2xy(inner, inter2) + polar2xy(opt*(1-width), inter2+0.5*np.pi),
        polar2xy(inner, inter2),
        polar2xy(inner, inter2),
        polar2xy(inner, inter2) + polar2xy(opt*(1-width), inter2-0.5*np.pi),
        polar2xy(inner, inter1) + polar2xy(opt*(1-width), inter1+0.5*np.pi),
        polar2xy(inner, inter1),
        polar2xy(inner, inter1),
        polar2xy(inner, inter1) + polar2xy(opt*(1-width), inter1-0.5*np.pi),
        polar2xy(inner, start) + polar2xy(opt*(1-width), start+0.5*np.pi),
        polar2xy(inner, start),
        polar2xy(radius, start),
    ]

    codes += [
        Path.LINETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.LINETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.LINETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.LINETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CLOSEPOLY,
    ]

    if ax is not None:
        path  = Path(verts, codes)
        patch = patches.PathPatch(path, facecolor=color, alpha=alpha,
                                  edgecolor=color, lw=LW)
        ax.add_patch(patch)

    return verts, codes


def chord_arc(start1, end1, start2, end2, radius=1.0, gap=0.03, pad=2,
              chordwidth=0.7, ax=None, color="r", cend="r", alpha=0.7,
              use_gradient=False, extent=360, directed=False):
    '''
    Draw a chord between two regions (arcs) of the chord diagram.

    Parameters
    ----------
    start1 : float (degree in 0, 360)
        Starting degree.
    end1 : float (degree in 0, 360)
        Final degree.
    start2 : float (degree in 0, 360)
        Starting degree.
    end2 : float (degree in 0, 360)
        Final degree.
    radius : float, optional (default: 1)
        External radius of the arc.
    gap : float, optional (default: 0)
        Distance between the arc and the beginning of the cord.
    chordwidth : float, optional (default: 0.2)
        Width of the chord.
    ax : matplotlib axis, optional (default: not plotted)
        Axis on which the chord should be plotted.
    color : valid matplotlib color, optional (default: "r")
        Color of the chord or of its beginning if `use_gradient` is True.
    cend : valid matplotlib color, optional (default: "r")
        Color of the end of the chord if `use_gradient` is True.
    alpha : float, optional (default: 0.7)
        Opacity of the chord.
    use_gradient : bool, optional (default: False)
        Whether a gradient should be use so that chord extremities have the
        same color as the arc they belong to.
    extent : float, optional (default : 360)
        The angular aperture, in degrees, of the diagram.
        Default is to use the whole circle, i.e. 360 degrees, but in some cases
        it can be useful to use only a part of it.
    directed : bool, optional (default: False)
        Whether the chords should be directed, ending in an arrow.

    Returns
    -------
    verts, codes : lists
        Vertices and path instructions to draw the shape.
    '''
    chordwidth2 = chordwidth

    dtheta1 = min((start1 - end2) % extent, (end2 - start1) % extent)
    dtheta2 = min((end1 - start2) % extent, (start2 - end1) % extent)

    start1, end1, verts, codes = initial_path(start1, end1, radius, chordwidth)

    if directed:
        if start2 > end2:
            start2, end2 = end2, start2

        start2 *= np.pi/180.
        end2   *= np.pi/180.

        tip = 0.5*(start2 + end2)
        asize = max(gap, 0.02)

        verts2 = [
            polar2xy(radius - asize, start2),
            polar2xy(radius, tip),
            polar2xy(radius - asize, end2)
        ]
    else:
        start2, end2, verts2, _ = initial_path(start2, end2, radius, chordwidth)

    chordwidth2 *= np.clip(0.4 + (dtheta1 - 2*pad) / (15*pad), 0.2, 1)

    chordwidth *= np.clip(0.4 + (dtheta2 - 2*pad) / (15*pad), 0.2, 1)

    rchord  = radius * (1-chordwidth)
    rchord2 = radius * (1-chordwidth2)

    verts += [polar2xy(rchord, end1), polar2xy(rchord, start2)] + verts2

    verts += [
        polar2xy(rchord2, end2),
        polar2xy(rchord2, start1),
        polar2xy(radius, start1),
    ]

    # update codes

    codes += [
        Path.CURVE4,
        Path.CURVE4,
    ]

    if directed:
        codes += [
            Path.CURVE4,
            Path.LINETO,
            Path.LINETO,
        ]
    else:
        codes += [
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            Path.LINETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            Path.LINETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            Path.LINETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
        ]

    codes += [
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
    ]

    if ax is not None:
        path = Path(verts, codes)

        if use_gradient:
            # find the start and end points of the gradient
            points, min_angle = None, None

            if dtheta1 < dtheta2:
                points = [
                    polar2xy(radius, start1),
                    polar2xy(radius, end2),
                ]

                min_angle = dtheta1
            else:
                points = [
                    polar2xy(radius, end1),
                    polar2xy(radius, start2),
                ]

                min_angle = dtheta1

            # make the patch
            patch = patches.PathPatch(path, facecolor="none",
                                      edgecolor="none", lw=LW)
            ax.add_patch(patch)  # this is required to clip the gradient

            # make the grid
            x = y = np.linspace(-1, 1, 100)
            meshgrid = np.meshgrid(x, y)

            gradient(points[0], points[1], min_angle, color, cend, meshgrid,
                     patch, ax, alpha)
        else:
            patch = patches.PathPatch(path, facecolor=color, alpha=alpha,
                                      edgecolor=color, lw=LW)

            idx = 16

            ax.add_patch(patch)

    return verts, codes


def self_chord_arc(start, end, radius=1.0, chordwidth=0.7, ax=None,
                   color=(1,0,0), alpha=0.7):
    start, end, verts, codes = initial_path(start, end, radius, chordwidth)

    rchord = radius * (1 - chordwidth)

    verts += [
        polar2xy(rchord, end),
        polar2xy(rchord, start),
        polar2xy(radius, start),
    ]

    codes += [
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
    ]

    if ax is not None:
        path  = Path(verts, codes)
        patch = patches.PathPatch(path, facecolor=color, alpha=alpha,
                                  edgecolor=color, lw=LW)
        ax.add_patch(patch)

    return verts, codes


#########################################################################
# File Name: scRNA_anno.py
# > Author: CaoYinghao
# > Mail: caoyinghao@gmail.com 
#########################################################################
#! /usr/bin/python

import sys
import argparse
import gzip
import os

import numpy as np
from numpy import asfarray,arange,minimum,abs,mat,sum,power,mean,array,std,log2,log10
import pandas as pd
from pandas import DataFrame,read_csv,ExcelWriter
from pickle import dump,load
from scipy.stats import fisher_exact
from scipy.sparse import coo_matrix

class Annotator(object):
    def __init__(self,foldchange,weight,
                 pvalue,tissue,species,
                 target,norefdb,MarkerDB,db,
                 noprint,input,output,source,cluster,fc,
                 outfmt,celltype,Gensymbol,list_tissue,cellrange):
        
        """
        Annotates marker genes using various databases and filtering criteria.

        Arguments:
            input: Input file for marker annotation (Only CSV format supported).
            output: Output file for marker annotation.
            db: Database for annotation (whole.db).
            source: Source of marker genes (cellranger, [seurat], [scanpy], [scran]).
            cluster: Only deal with one cluster of marker genes (all, [1], [1, 2, 3], [...]).
            markerdb: User-defined marker database in table format with two columns. First column as Cellname, second column as Genename.
            foldchange: Fold change threshold for marker filtering (2.0).
            pvalue: P-value threshold for marker filtering (0.05).
            weight: Weight threshold for marker filtering from cellranger v1.0 results (100).
            species: Species for annotation. Only used for cellmarker database ('Human', ['Mouse']).
            tissue: Tissue for annotation. Only used for cellmarker database. Multiple tissues should be separated by commas.
                In linux platform: ('All', ['Bone marrow'], ['Bone marrow, Brain, Blood'], [...]).
                In windows platform: ("All", ["Bone marrow"], ["Bone marrow, Brain, Blood"], [...]).
            outfmt: Output file format for marker annotation (ms-excel, [txt]).
            celltype: Cell type for annotation (normal, [cancer]).
            target: Target to annotation class in database (cellmarker, [cancersea]).
            gensymbol: Using gene symbol ID instead of ensembl ID in input file for calculation.
            norefdb: Only using user-defined marker database for annotation.
            noprint: Do not print any detail results.
            list_tissue: List tissue names in database.
            
            """
        
        #self = args
        self.input=input
        self.output=output
        self.source=source 
        self.cluster=cluster 
        self.fc = fc 
        self.db=db
        self.outfmt = outfmt 
        self.celltype = celltype
        self.Gensymbol = Gensymbol
        self.list_tissue =list_tissue
        self.foldchange = float(foldchange)
        self.weight = float(weight)
        self.pvalue = float(pvalue)
        self.tissue=tissue
        self.species=species 
        self.target=target
        self.norefdb=norefdb 
        self.MarkerDB=MarkerDB
        self.noprint=noprint
        self.cellrange=cellrange
        
        pass

    @staticmethod
    def do_fisher_test(x,num1,num2):
        """do fisher-exact test for go annotation"""
        fy = x['gene_num']
        by = x['othergene_num']
        o,p = fisher_exact([[fy,num1 - fy],[by,num2 - by]],alternative="greater")
        return p
        pass

    @staticmethod
    def do_sig_tag(x):
        if x <= 0.001:
            return "***"
        elif x <= 0.01:
            return "**"
        elif x <= 0.05:
            return "*"
        else:
            return "-"

    @staticmethod
    def p_adjust_bh(p):
        """Benjamini-Hochberg p-value correction for multiple hypothesis testing."""
        p = asfarray(p)
        by_descend = p.argsort()[::-1]
        by_orig = by_descend.argsort()
        steps = float(len(p)) / arange(len(p), 0, -1)
        q = minimum(1, minimum.accumulate(steps * p[by_descend]))
        return q[by_orig]

    @staticmethod
    def to_output(h_values,wb,outtag,cname,title):
        if outtag.lower() == "ms-excel":
            h_values.to_excel(wb,sheet_name = "Cluster " + cname + " " + title,index=False)
        else:
            def clean_non_printable_chars(value):
                """移除字符串中的非打印字符"""
                if isinstance(value, str):
                    return ''.join(c for c in value if c.isprintable())
                return value
            h_values = h_values.applymap(clean_non_printable_chars)
            h_values.to_csv(wb,sep="\t",quotechar = "\t",index=False,header=False,encoding='gbk')

    @staticmethod
    def translate_go(name="go.obo"):
        """extract go annotation dicts from obo"""
        ids = []
        names = []
        cls = []
        infile = None
        if name.endswith(".gz"):
            infile = gzip.open(name,"rt")
        else:
            infile = open(name,"rt")
        for line in infile:
            if line.strip().startswith("id:"):
                ids.append(line.strip()[4:])
            if line.strip().startswith("name:"):
                names.append(line.strip()[6:])
            if line.strip().startswith("namespace:"):
                cls.append(line.strip()[11:])
        infile.close()
        return dict(zip(ids,names))


    def do_go_annotation(self,gof,fore,back,cname,gtype):
        """return go annotation with significance tag"""

        if self.year == 2024:
            fil = gof['2'].map(lambda value: len(set([value]) & fore) > 0)
            fgnames = gof[fil].groupby(by="4")["2"].unique()
            bfil = gof["2"].map(lambda value: len(set([value]) & back) > 0)
            bgnames = gof[bfil].groupby(by="4")["2"].unique()
        elif self.year == 2023:
            fil = gof[2].map(lambda value: len(set([value]) & fore) > 0)
            fgnames = gof[fil].groupby(by=4)[2].unique()
            bfil = gof[2].map(lambda value: len(set([value]) & back) > 0)
            bgnames = gof[bfil].groupby(by=4)[2].unique()
        else:
            fil = gof[2].map(lambda value: len(set([value]) & fore) > 0)
            fgnames = gof[fil].groupby(by=4)[2].unique()
            bfil = gof[2].map(lambda value: len(set([value]) & back) > 0)
            bgnames = gof[bfil].groupby(by=4)[2].unique()
            
        dat = DataFrame({"genes":fgnames,"othergenes":bgnames})
        num1 = len(fore)
        num2 = len(back)
        dat = dat.dropna(axis=0)
        dat['gene_num'] = dat['genes'].map(lambda x:x.size)
        dat['othergene_num'] = dat['othergenes'].map(lambda x:x.size)
        if dat.size ==0:
            return dat

        dat['p-value'] =  dat.apply(Annotator.do_fisher_test,axis=1,args=(num1,num2))
        used = dat.loc[:,dat.columns]   #.head(5)
        used['ids'] = used.index
        used['q-value'] = Annotator.p_adjust_bh(used['p-value'])
        used['sig'] = used['q-value'].map(Annotator.do_sig_tag)
        used['go_name'] = used['ids'].map(lambda x:self.gos[x])
        outs = used.sort_values(by='p-value')[['ids','gene_num','othergene_num','p-value','q-value',"sig",'go_name']]
        if self.noprint == False:
            print(outs.head(5).to_string(index=False))
        if self.output:
            outs['cluster'] = cname
            outs['go_class'] = gtype
        return outs
        pass

    def print_class(self,h_values,cname):
        """print cell predictions with scores."""
        o = ""
        titlebar = "-" * 60 + "\n"
        #print(h_values)
        #print(h_values.size)
        #if h_values is not None:
        #    print(h_values.size)
        if h_values is None:
            if self.noprint:
                return "E",None,"-","-","-"
            o += titlebar
            o += "{0:<10}{1:^30}{2:<10}".format("Type","Cell Type","Score")
            o += "\n" + "-"*60 + "\n"
            o += "{0:<10}{1:^30}{2:<10}".format("-","-","-")
            o += "\n" + titlebar
            return "E",None,"-","-","-"
        elif h_values.size == 0:
            if self.noprint:
                return "N",None,"-","-","-"
            o += titlebar
            o += "{0:<10}{1:^30}{2:<10}".format("Type","Cell Type","Score")
            o += "\n" + "-"*60 + "\n"
            o += "{0:<10}{1:^30}{2:<10}".format("-","-","-")
            o += "\n" + titlebar
            return "N",None,"-","-","-"
        elif h_values.size == 3:
            if self.noprint:
                return "Good",o,h_values.values[0][0],h_values.values[0][1],"-"
            o += titlebar
            o += "{0:<10}{1:^30}{2:<10}{3:<5}".format("Type","Cell Type","Score","Times")
            o += "\n" + "-"*60 + "\n"
            o += "{0:<10}{1:^30}{2:<10.4f}".format("Good",h_values.values[0][0],h_values.values[0][1])
            o += "\n" + titlebar
            return "Good",o,h_values.values[0][0],h_values.values[0][1],"-"
            pass
        elif h_values.size == 2:
            if self.noprint:
                return "Good",o,h_values.values[0][0],h_values.values[0][1],"-"
            o += titlebar
            o += "{0:<10}{1:^30}{2:<10}{3:<5}".format("Type","Cell Type","Score","Times")
            o += "\n" + "-"*60 + "\n"
            o += "{0:<10}{1:^30}{2:<10.4f}".format("Good",h_values.values[0][0],h_values.values[0][1])
            o += "\n" + titlebar
            return "Good",o,h_values.values[0][0],h_values.values[0][1],"-"
            pass
        elif float(h_values.iloc[0,1])/float(h_values.iloc[1,1]) >= 2 or float(h_values.iloc[1,1] < 0):
            times = np.abs(float(h_values.iloc[0,1])/float(h_values.iloc[1,1]))
            if self.noprint:
                return "Good",o,h_values.values[0][0],h_values.values[0][1],times
            o += titlebar
            o += "{0:<10}{1:^30}{2:<10}{3:<5}".format("Type","Cell Type","Score","Times")
            o += "\n" + titlebar
            o += "{0:<10}{1:^30}{2:<10.4f}{3:<5.1f}".format("Good",h_values['Cell Type'].values[0],h_values['Z-score'].values[0],times)
            o += "\n" + titlebar
            return "Good",o,h_values['Cell Type'].values[0],h_values['Z-score'].values[0],times
            pass
        else:
            times = np.abs(float(h_values.iloc[0,1])/float(h_values.iloc[1,1]))
            if self.noprint:
                return "?",o,str(h_values['Cell Type'].values[0]) + "|" + str(h_values['Cell Type'].values[1]),str(h_values['Z-score'].values[0]) + "|" + str(h_values['Z-score'].values[1]),times
            o += titlebar
            o += "{0:<10}{1:^30}{2:<10}{3:<5}".format("Type","Cell Type","Score","Times")
            o += "\n" + titlebar
            o += "{0:<10}{1:^30}{2:<10.4f}{3:<5.1f}".format("?",h_values['Cell Type'].values[0],h_values['Z-score'].values[0],times)
            o += "\n" + titlebar
            o += "{0:<10}{1:^29}({2:<.4f})".format("","("+h_values['Cell Type'].values[1]+")",h_values['Z-score'].values[1])
            o += "\n" + titlebar
            return "?",o,str(h_values['Cell Type'].values[0]) + "|" + str(h_values['Cell Type'].values[1]),str(h_values['Z-score'].values[0]) + "|" + str(h_values['Z-score'].values[1]),times
        pass

    def deal_with_badtype(self,cname,other_gene_names,colnames):
        """go annotation need to be performed"""
        if len(self.human_gofs) != 0:
            fset = set()
            bset = set()
            if colnames is None:
                print("!WARNING(go processing):Zero gene sets found for the cluster",cname)
                print("!WARNING(go processing):Change the threshold and try again?")
                return 
            for c in colnames:
                if c in self.ensem_hgncs:
                    fset.add(self.ensem_hgncs[c])
                else:
                    fset.add(c)
            for c in other_gene_names:
                if c in self.ensem_hgncs:
                    bset.add(self.ensem_hgncs[c])
                else:
                    bset.add(c)
            if len(fset) == 0:
                print("!WARNING(go processing):Zero gene sets found for the cluster",cname)
                print("!WARNING(go processing):Change the threshold and try again?")
                return 
            if len(bset) == 0:
                print("!WARNING(go processing):Zero gene sets found for other clusters")
                print("!WARNING(go processing):Change the threshold and try again?")
                return 
            names = ["Function","Component","Process"]
            if self.noprint == False:
                print("Go Enrichment analysis:","Group1:",len(fset),"Group2:",len(bset))
            if len(fset) > 0 and len(bset) > 0:
                all_outs = DataFrame()
                for i,f in enumerate(self.human_gofs):
                    o = " ".join([">"*30,names[i], "<"*30])
                    if self.noprint == False:
                        print(o)
                    outs = self.do_go_annotation(f,fset,bset,cname,names[i][0])
                    if outs.size == 0:continue
                    if all_outs.size == 0:
                        all_outs = outs
                    else:
                        all_outs = pd.concat([all_outs,outs])
                    if self.noprint == False:
                        print()
                if self.output:
                    Annotator.to_output(all_outs,self.wbgo,self.outfmt,cname,"GO")



    def calcu_cellranger_group(self,expfile,hgvc=False):
        """deal with cellranger input matrix"""
        exps = read_csv(expfile)
        columns = exps.columns

        pre,suf,suf1 ="Cluster "," UMI counts/cell"," Weight"
        fid = "Gene Name" if hgvc == True else "Gene ID"
        gcol = "gene" if hgvc == True else "ensemblID"
        ccol = "cellName"

        if self.target.lower() not in ["cancersea","cellmarker","panglaodb"]:
            print("Error target : -t, --target,(cellmarker,[cancersea,panglaodb])")
            sys.exit(0)

        if self.target.lower() == "cancersea":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"
        if self.target.lower() == "panglaodb":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"


        abs_tag = False

        cnum = int(len(exps.columns) / 2 - 1)
        ver_tag = "V1"
        pname = ""

        if "Feature ID" in columns: # v3
            fid = "Feature Name" if hgvc == True else "Feature ID"
            pre,suf,suf1 ="Cluster "," Mean Counts"," Log2 fold change"
            cnum = int((len(exps.columns)-2) / 3)
            pname = " Adjusted p value"
            self.weight = self.foldchange
            ver_tag = "V3"
        elif "Cluster 1 Mean UMI Counts" in columns: # v2
            fid = "Gene Name" if hgvc == True else "Gene ID"
            pre,suf,suf1 ="Cluster "," Mean Counts"," Log2 fold change"
            cnum = int((len(exps.columns)-2) / 3)
            self.weight = self.foldchange
            pname = " Adjusted p value"
            ver_tag = "V2"
        outs = []

        self.wb = self.wbgo = None
        if self.output:
            if self.outfmt.lower() == "ms-excel":
                if not self.output.endswith(".xlsx") and (not self.output.endswith(".xls")):
                    self.output += ".xlsx"
                self.wb = ExcelWriter(self.output)
                self.wbgo = self.wb
            elif self.outfmt.lower() == "txt":
                self.wb = open(self.output,"w", encoding='utf-8')
                self.wb.write("Cell Type\tZ-score\tCluster\n")
                self.wbgo = open(self.output + ".go","w", encoding='utf-8')
                self.wbgo.write('ids\tgene_num\tothergene_num\tp-value\tq-value\tsig\tname\tcluster\tgo_class\n')
            else:
                print("Error output format: -m, --outfmt,(ms-excel,[txt])")
                sys.exit(0)

        

        for i in range(1,cnum+1):
            cname = str(i)
            if self.cluster != "all":
                if self.cluster.find(",") > -1:
                    sets = self.cluster.split(",")
                    if cname not in sets:
                        continue
                else:
                    if cname != self.cluster:
                        continue
            #if i != 1 :continue
            o = " ".join(["#"*30,"Cluster",cname, "#"*30]) + "\n"
            if self.noprint == False:
                print(o)
            ptitle = pre + cname + pname
            ltitle = pre + cname + suf1
            if ltitle not in exps.columns:
                print(ltitle,"column not in the input table!")
                sys.exit(0)

            newexps = None
            if ver_tag == "V1":
                newexps = exps[exps[ltitle]>=self.weight]
            else:
                newexps = exps[(exps[ltitle]>=self.weight) & (exps[ptitle] <= self.pvalue)]
            #print(newexps.shape)
            h_values,colnames = self.get_cell_matrix(newexps,ltitle,fid,gcol,ccol,abs_tag)
            #print(newexps)
            print("Cluster " + cname + " Gene number:",newexps[fid].unique().shape[0])
            if h_values is None:
                t,o_str,c,v,times = self.print_class(h_values,cname)
                outs.append([cname,t,c,v,times])
                if self.noprint == False:
                    print(o_str)
                continue
            h_values['Cluster'] = cname
            if self.output:
                Annotator.to_output(h_values,self.wb,self.outfmt,cname,"Cell Type")


            t,o_str,c,v,times = self.print_class(h_values,cname)
            outs.append([cname,t,c,v,times])
            if self.noprint == False:
                print(o_str)
            other_gene_names = set()
            for j in range(1,cnum + 1):
                if i ==j :continue
                jtitle = pre + str(j) + suf1
                otherexps = None
                if ver_tag == "V1":
                    otherexps = exps[exps[jtitle]>=self.weight]
                else:
                    otherexps = exps[(exps[jtitle]>=self.weight) & (exps[ptitle] <= self.pvalue)]
                if self.target.lower() == "cancersea":
                    tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.smarkers,fid,gcol,ccol,"other")
                    other_gene_names |= set(tcolnames)
                elif self.target.lower() == "cellmarker":
                    tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.cmarkers,fid,gcol,ccol,"other")
                    if not trownames:
                        #print("WARNING3:Zero gene sets found for the cluster" + str(j))
                        #print("WARNING3:Change the threshold and try again?")
                        continue
                    other_gene_names |= set(tcolnames)
                elif self.target.lower() == "panglaodb":
                    tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.pmarkers,fid,gcol,ccol,"other")
                    if not trownames:
                        #print("WARNING3:Zero gene sets found for the cluster" + str(j))
                        #print("WARNING3:Change the threshold and try again?")
                        continue
                    other_gene_names |= set(tcolnames)
            #print("Other Gene number:",len(other_gene_names))
            self.deal_with_badtype(cname,other_gene_names,colnames)
        if self.output:
            self.wb.close()
            self.wbgo.close()
        if self.noprint == False:
            print("#"*80 + "\n")
        return outs

    def calcu_seurat_group(self,expfile,hgvc=False):
        """deal with seurat input matrix"""
        exps = read_csv(expfile)
        pre,suf,suf1 ="avg_logFC"," UMI counts/cell",""
        fid = "gene"
        pname = "p_val_adj"
        assert fid in exps.columns, 'No "gene" column. Wrong format? Seurat, Scanpy, Scran or Cellranger?'
        exps[fid] = exps[fid].str.replace("\.\d+","")
        cluster = "cluster"
        gcol = "gene" if hgvc == True else "ensemblID"
        ccol = "cellName"

        if self.target.lower() not in ["cancersea","cellmarker","panglaodb"]:
            print("Error target : -t, --target,(cellmarker,[cancersea,panglaodb])")
            sys.exit(0)

        if self.target.lower() == "cancersea":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"
        if self.target.lower() == "panglaodb":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"

        cnum = list(exps[cluster].unique())
        abs_tag = True
        outs = []
        self.wb = self.wbgo = None
        if self.output:
            if self.outfmt.lower() == "ms-excel":
                if not self.output.endswith(".xlsx") and (not self.output.endswith(".xls")):
                    self.output += ".xlsx"
                self.wb = ExcelWriter(self.output)
                self.wbgo = self.wb
            elif self.outfmt.lower() == "txt":
                self.wb = open(self.output,"w", encoding='utf-8')
                if self.target == "cancersea":
                    self.wb.write("Cell Type\tZ-score\tCluster\n")
                else:
                    self.wb.write("Cell Type\tZ-score\tCluster\n")
                self.wbgo = open(self.output + ".go","w", encoding='utf-8')
                self.wbgo.write('ids\tgene_num\tothergene_num\tp-value\tq-value\tsig\tname\tcluster\tgo_class\n')
            else:
                print("Error output format: -m, -outfmt,(ms-excel,[txt])")
                sys.exit(0)

        for i in cnum:
            cname = str(i)
            if self.cluster != "all":
                if self.cluster.find(",") > -1:
                    sets = self.cluster.split(",")
                    if cname not in sets:
                        continue
                else:
                    if cname != self.cluster:
                        continue
            o = " ".join(["#"*30,"Cluster",cname, "#"*30]) + "\n"
            if self.noprint == False:
                print(o)
            ltitle = pre
            ptitle = pname
            if ltitle not in exps.columns:
                print(ltitle,"column not in the input table!")
                sys.exit(0)
            newexps = exps[(exps[cluster] == i) & (exps[ltitle]>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #newexps = exps[(exps[cluster] == i) & (abs(exps[ltitle])>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #print(newexps)
            #print(newexps)

            h_values,colnames = self.get_cell_matrix(newexps,ltitle,fid,gcol,ccol,abs_tag)
            print("Cluster " + cname + " Gene number:",newexps['gene'].unique().shape[0])
            #print(colnames)
            #for x in newexps['gene'].unique():
            #    print(x)
            #exit()
            if self.output:
                h_values['Cluster'] = cname
                Annotator.to_output(h_values,self.wb,self.outfmt,cname,"Cell Type")

            #print(h_values)
            t,o_str,c,v,times = self.print_class(h_values,cname)
            outs.append([cname,t,c,v,times])
            if self.noprint == False:
                print(o_str)

            otherexps = exps[(exps[cluster] != i) & (exps[ltitle]>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #otherexps = exps[(exps[cluster] != i) & (abs(exps[ltitle])>=self.foldchange) & (exps[ptitle] <= self.pvalue)]

            if self.target.lower() == "cellmarker":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.cmarkers,fid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            elif self.target.lower() == "cancersea":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.smarkers,fid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            elif self.target.lower() == "panglaodb":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.pmarkers,fid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            print("Other Gene number:",len(other_gene_names))
        if self.output:
            self.wb.close()
            self.wbgo.close()
        if self.noprint == False:
            print("#"*80 + "\n")
        return outs

    def calcu_scanpy_group(self,expfile,hgvc=False):
        """deal with scanpy input matrix"""
        exps = read_csv(expfile,index_col=0)
        cnum = set()
        pname = "p"
        pre = "l"
        fid = "n"
        for c in exps.columns:
            k,v = c.split("_")
            cnum.add(k)
            if v.startswith("p"):
                pname = v
            elif v.startswith("n"):
                rfid = v
            elif v.startswith("l"):
                pre = v
        
        #pre,suf,suf1 ="avg_logFC"," UMI counts/cell",""
        #fid = "gene"
        #pname = "p_val_adj"
        #assert fid in exps.columns, 'No "gene" column. Wrong format? Scanpy, Seurat or Cellranger?'
        #exps[fid] = exps[fid].str.replace("\.\d+","")
        #cluster = "cluster"

        ###MarkerBase
        gcol = "gene" if hgvc == True else "ensemblID"
        ccol = "cellName"

        if self.target.lower() not in ["cancersea","cellmarker","panglaodb"]:
            print("Error target : -t, --target,(cellmarker,[cancersea,panglaodb])")
            sys.exit(0)

        if self.target.lower() == "cancersea":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"
        if self.target.lower() == "panglaodb":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"

        #cnum = list(exps[cluster].unique())
        abs_tag = True
        outs = []
        self.wb = self.wbgo = None
        if self.output:
            if self.outfmt.lower() == "ms-excel":
                if not self.output.endswith(".xlsx") and (not self.output.endswith(".xls")):
                    self.output += ".xlsx"
                self.wb = ExcelWriter(self.output)
                self.wbgo = self.wb
            elif self.outfmt.lower() == "txt":
                self.wb = open(self.output,"w", encoding='utf-8')
                if self.target == "cancersea":
                    self.wb.write("Cell Type\tZ-score\tCluster\n")
                else:
                    self.wb.write("Cell Type\tZ-score\tCluster\n")
                self.wbgo = open(self.output + ".go","w", encoding='utf-8')
                self.wbgo.write('ids\tgene_num\tothergene_num\tp-value\tq-value\tsig\tname\tcluster\tgo_class\n')
            else:
                print("Error output format: -m, -outfmt,(ms-excel,[txt])")
                sys.exit(0)

        for i in list(sorted(cnum)):
            cname = str(i)
            if self.cluster != "all":
                if self.cluster.find(",") > -1:
                    sets = self.cluster.split(",")
                    if cname not in sets:
                        continue
                else:
                    if cname != self.cluster:
                        continue
            o = " ".join(["#"*30,"Cluster",cname, "#"*30]) + "\n"
            if self.noprint == False:
                print(o)
            ltitle = cname + "_" + pre
            fid = cname + "_" + rfid
            ptitle = cname + "_" + pname
            if ltitle not in exps.columns:
                print(ltitle,"column not in the input table!")
                sys.exit(0)
            newexps = exps[[fid,ltitle,ptitle]][(exps[ltitle]>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #newexps = exps[(exps[cluster] == i) & (abs(exps[ltitle])>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #print(newexps)
            #print(newexps)

            h_values,colnames = self.get_cell_matrix(newexps,ltitle,fid,gcol,ccol,abs_tag)
            print("Cluster " + cname + " Gene number:",newexps[fid].unique().shape[0])
            #print(colnames)
            #for x in newexps[fid].unique():
            #    print(x)
            #exit()
            if self.output:
                h_values['Cluster'] = cname
                Annotator.to_output(h_values,self.wb,self.outfmt,cname,"Cell Type")

            #print(h_values)
            #exit()
            t,o_str,c,v,times = self.print_class(h_values,cname)
            outs.append([cname,t,c,v,times])
            if self.noprint == False:
                print(o_str)

            otherexps = None
            ofid = 'o_n'
            oltitle = 'o_l'
            optitle = 'o_p'
            for j in list(sorted(cnum)):
                oname = str(j)
                if oname == cname:continue
                tltitle = oname + "_" + pre
                tfid = oname + "_" + rfid
                tptitle = oname + "_" + pname
                tempexps = exps[[tfid,tltitle,tptitle]][(exps[tltitle]>=self.foldchange) & (exps[tptitle] <= self.pvalue)]
                tempexps.columns = [ofid,oltitle,optitle]
                if otherexps is None:
                    otherexps = tempexps
                else:
                    otherexps = pd.concat([otherexps,tempexps])
            #otherexps = exps[(exps[cluster] != i) & (abs(exps[ltitle])>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #print(otherexps)
            #exit()

            if self.target.lower() == "cellmarker":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.cmarkers,ofid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            elif self.target.lower() == "cancersea":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.smarkers,ofid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            elif self.target.lower() == "panglaodb":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.pmarkers,ofid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            print("Other Gene number:",len(other_gene_names))
        if self.output:
            self.wb.close()
            self.wbgo.close()
        if self.noprint == False:
            print("#"*80 + "\n")
        return outs

    def calcu_scran_group(self,expfile,hgvc=False):
        """deal with scran input matrix"""
        exps = read_csv(expfile)
        if exps.columns[0] == "Unnamed: 0":
            exps.rename(columns={"Unnamed: 0":"gene"},inplace=True)
        cnum = set()
        pname = "p.value"
        pre = "LFC"
        fid = "gene"
        for c in exps.columns:
            if c.startswith("LFC"):
                k,v = c.split("_")
                cnum.add(v)
        
        #pre,suf,suf1 ="avg_logFC"," UMI counts/cell",""
        #fid = "gene"
        #pname = "p_val_adj"
        #assert fid in exps.columns, 'No "gene" column. Wrong format? Scanpy, Seurat or Cellranger?'
        #exps[fid] = exps[fid].str.replace("\.\d+","")
        #cluster = "cluster"

        ###MarkerBase
        gcol = "gene" if hgvc == True else "ensemblID"
        ccol = "cellName"

        if self.target.lower() not in ["cancersea","cellmarker","panglaodb"]:
            print("Error target : -t, --target,(cellmarker,[cancersea,panglaodb])")
            sys.exit(0)

        if self.target.lower() == "cancersea":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"

        if self.target.lower() == "panglaodb":
            gcol = "gene" if hgvc == True else "ensemblID"
            ccol = "name"

        #cnum = list(exps[cluster].unique())
        abs_tag = True
        outs = []
        self.wb = self.wbgo = None
        if self.output:
            if self.outfmt.lower() == "ms-excel":
                if not self.output.endswith(".xlsx") and (not self.output.endswith(".xls")):
                    self.output += ".xlsx"
                self.wb = ExcelWriter(self.output)
                self.wbgo = self.wb
            elif self.outfmt.lower() == "txt":
                self.wb = open(self.output,"w", encoding='utf-8')
                if self.target == "cancersea":
                    self.wb.write("Cell Type\tZ-score\tCluster\n")
                else:
                    self.wb.write("Cell Type\tZ-score\tCluster\n")
                self.wbgo = open(self.output + ".go","w", encoding='utf-8')
                self.wbgo.write('ids\tgene_num\tothergene_num\tp-value\tq-value\tsig\tname\tcluster\tgo_class\n')
            else:
                print("Error output format: -m, -outfmt,(ms-excel,[txt])")
                sys.exit(0)

        for i in list(sorted(cnum)):
            cname = str(i)
            if self.cluster != "all":
                if self.cluster.find(",") > -1:
                    sets = self.cluster.split(",")
                    if cname not in sets:
                        continue
                else:
                    if cname != self.cluster:
                        continue
            o = " ".join(["#"*30,"Cluster",cname, "#"*30]) + "\n"
            if self.noprint == False:
                print(o)
            ltitle = pre + "_" + cname
            ptitle = pname + "_" + cname
            if ltitle not in exps.columns:
                print(ltitle,"column not in the input table!")
                sys.exit(0)
            newexps = exps[[fid,ltitle,ptitle]][(exps[ltitle]>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #newexps = exps[(exps[cluster] == i) & (abs(exps[ltitle])>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #print(newexps)
            #print(newexps)
            #continue

            h_values,colnames = self.get_cell_matrix(newexps,ltitle,fid,gcol,ccol,abs_tag)
            print("Cluster " + cname + " Gene number:",newexps[fid].unique().shape[0])
            #print(colnames)
            #for x in newexps[fid].unique():
            #    print(x)
            #exit()
            if self.output:
                h_values['Cluster'] = cname
                Annotator.to_output(h_values,self.wb,self.outfmt,cname,"Cell Type")

            #print(h_values)
            #exit()
            t,o_str,c,v,times = self.print_class(h_values,cname)
            outs.append([cname,t,c,v,times])
            if self.noprint == False:
                print(o_str)

            otherexps = None
            ofid = 'o_n'
            oltitle = 'o_l'
            optitle = 'o_p'
            for j in list(sorted(cnum)):
                oname = str(j)
                if oname == cname:continue
                tltitle = pre + "_" + oname
                tfid = fid 
                tptitle = pname + "_" + oname
                tempexps = exps[[tfid,tltitle,tptitle]][(exps[tltitle]>=self.foldchange) & (exps[tptitle] <= self.pvalue)]
                tempexps.columns = [ofid,oltitle,optitle]
                if otherexps is None:
                    otherexps = tempexps
                else:
                    otherexps = pd.concat([otherexps,tempexps])
            #otherexps = exps[(exps[cluster] != i) & (abs(exps[ltitle])>=self.foldchange) & (exps[ptitle] <= self.pvalue)]
            #print(otherexps)
            #exit()

            if self.target.lower() == "cellmarker":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.cmarkers,ofid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            elif self.target.lower() == "cancersea":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.smarkers,ofid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            elif self.target.lower() == "panglaodb":
                tfc,trownames,trownum,tcolnames,tcolnum = self.get_cell_gene_names(otherexps,self.pmarkers,ofid,gcol,ccol,'other')
                if not trownames:continue
                other_gene_names = set(tcolnames)
                self.deal_with_badtype(cname,other_gene_names,colnames)
            print("Other Gene number:",len(other_gene_names))
        if self.output:
            self.wb.close()
            self.wbgo.close()
        if self.noprint == False:
            print("#"*80 + "\n")
        return outs


    def get_exp_matrix_loop(self,exps,ltitle,fid,colnames,rownames,cell_matrix,usertag,abs_tag = True):
        """format the cell_deg_matrix and calculate the zscore of certain cell types."""

        #filter gene expressed matrix according to the markers
        gene_exps = exps.loc[:,[fid,ltitle]][exps[fid].isin(colnames)]

        gene_matrix = mat(gene_exps.sort_values(fid)[ltitle]).T
        gene_matrix = gene_matrix * np.mean(gene_matrix) ### / np.min(gene_matrix))

        if gene_matrix.shape[0] != cell_matrix.shape[1]:
            #print(gene_matrix.shape,cell_matrix.shape)
            #print(len(gene_exps[fid].unique()))
            #print(gene_matrix)
            #print(cell_matrix)
            print("Error for inconsistent gene numbers, please check your expression csv for '" + fid + "'")
            return None
        
        nonzero = np.matrix(np.count_nonzero(cell_matrix,axis=1)).T
        #gene_matrix = np.ones_like(gene_matrix)
        cell_deg_matrix = cell_matrix * gene_matrix

        #print("cell",cell_matrix)
        #print("gene",gene_matrix)
        #print(colnames)
        #print(rownames)

        #print(rownames)
        #exit()
        #print(gene_matrix)
        #print(cell_deg_matrix)
        #print(type(rownames))
        #a1 = "Natural killer T (NKT) cell"
        #b1 = "T cell"
        #a1 = "Macrophage"
        #b1 = "Monocyte"

        #a1 = "Mesenchymal stem cell"
        #b1 = "Fibroblast"
        #mar = cell_matrix[np.array(rownames) == a1]
        #mon = cell_matrix[np.array(rownames) == b1]
        #marz = nonzero[np.array(rownames) == a1]
        #monz = nonzero[np.array(rownames) == b1]
        #print(len(mar[np.nonzero(mar)]),len(mon[np.nonzero(mon)]))

        #print(log2(marz),log2(monz))

        #print(mar)
        #print(mon)
        #print(marz,monz)
        #print(cell_matrix,cell_matrix.shape,gene_matrix.shape)

        #print(np.std(cell_matrix,axis=1))
        #print(cell_matrix.shape,cell_deg_matrix.shape)
        wstd = np.matrix(np.std(cell_matrix,axis=1)).T
        #print(wstd.shape,wstd,nonzero)
        if usertag:
            cell_deg_matrix = np.matrix(np.array(cell_deg_matrix))
        else:
            if (wstd.shape == np.ones_like(wstd)).all:
                wstd = [[1]]
            if (nonzero == np.ones_like(nonzero)).all:
                cell_deg_matrix = np.matrix(np.array(cell_deg_matrix) * np.array(wstd))
            else:
                cell_deg_matrix = np.matrix(np.array(cell_deg_matrix) * np.array(log2(nonzero)) * np.array(wstd))

        out = DataFrame({"Z-score":cell_deg_matrix.A1},index=rownames)
        out.sort_values(['Z-score'],inplace=True,ascending=False)
        #out.to_csv("wei.sco",sep="\t")
        #print(cell_deg_matrix,wstd,log2(nonzero))

        if abs_tag:
            out['Z-score'] = abs(out['Z-score'])
        else:
            out = out[out['Z-score'] > 0]

        #print(out)
        if (out.shape[0] > 1):
            out['Z-score'] = (out['Z-score'] - mean(out['Z-score']))/std(out['Z-score'],ddof=1)
        #print(out)

        return out


    def get_cell_gene_names(self,exps,markers,fid,gcol,ccol,tag):
        """find expressed markers according to the markers and expressed matrix."""
        #print(fid)
        whole_gsets = set(exps[fid])
        if self.target.lower() == "cancersea" or self.target.lower() == "panglaodb":
            #whole_fil = markers['EnsembleID'].isin(whole_gsets)
            markers['weight'] = 1
            if self.Gensymbol == True:
                markers[gcol] = markers['GeneName']
            else:
                markers[gcol] = markers['EnsembleID']
        whole_fil = markers[gcol].isin(whole_gsets)

        fc = markers[[ccol,gcol,'weight']][whole_fil]
        #print(whole_gsets,exps[fid],exps)
        #print(markers[gcol])
        #print(fc)
        #print(list(fc['cellName'].unique()),ccol,gcol)
        #print(whole_fil.unique())
        #exit()
        #print(markers,markers.columns)
        #print(fc)
        if fc.shape[0] == 0:
            if tag != "other":
                print("!WARNING3:Zero marker sets found, type:" + tag)
                print("!WARNING3:Change the threshold or tissue name and try again?")
                print("!WARNING3:EnsemblID or GeneID,try '-E' command?")
            return fc,None,None,whole_gsets,None
        #print("helll")
        fc.columns = [ccol,gcol,'c']
        fc.set_index([ccol,gcol])
        newfc = fc.groupby([ccol,gcol]).sum()
        #print(newfc)
        #if newfc.shape[0] <1:
        #    print(newfc.shape)
        #    print(fc)
        #    print(exps)
        names = newfc.index
        #print(names)
        #print(names)
        newfc['c1'] = names
        newfc[gcol] = newfc['c1'].apply(lambda x:x[1])
        newfc[ccol] = newfc['c1'].apply(lambda x:x[0])
        newfc.drop(['c1'],inplace=True,axis=1)
        newfc.reset_index(drop=True,inplace=True)
        #print(newfc)
        #exit()
        newfc['c'] = log2(newfc['c'] + 0.05) # * np.min(newfc['c'])
        fc = newfc
        #print("hello")
        #newfc.to_csv("wei.cls",sep="\t")
        #exit()
        #print(fc['c'][fc['c'] != 0])

        rownames = sorted(set(fc[ccol].unique()))
        rownum = len(rownames)
        colnames = sorted(set(fc[gcol].unique()))
        colnum = len(colnames)
        #print(fc.shape,fc)
        return fc,rownames,rownum,colnames,colnum

    def get_user_cell_gene_names(self,exps,fid,gcol,ccol,tag):
        """find expressed markers according to the user markers and expressed matrix."""
        #print(self.usermarkers)
        self.usermarkers.columns = [ccol,gcol,'weight']
        whole_gsets = set(exps[fid])
        whole_fil = self.usermarkers[gcol].isin(whole_gsets)

        fc = self.usermarkers[[ccol,gcol,'weight']][whole_fil]
        if fc.shape[0] == 0:
            if tag != "other":
                print("!WARNING3:Zero marker sets found, type:" + tag)
                print("!WARNING3:Change the threshold or tissue name and try again?")
                print("!WARNING3:EnsemblID or GeneID,try '-E' command?")
            return fc,None,None,whole_gsets,None
        fc.columns = [ccol,gcol,'c']
        fc.set_index([ccol,gcol])
        #print("FC",fc)
        #print("ENSG00000105369" in whole_gsets)

        newfc = fc.groupby([ccol,gcol]).sum()
        #if newfc.shape[0] <1:
        #    print(newfc.shape)
        #    print(fc)
        #    print(exps)
        names = newfc.index
        #print(names)
        #print(names)
        newfc['c1'] = names
        newfc[gcol] = newfc['c1'].apply(lambda x:x[1])
        newfc[ccol] = newfc['c1'].apply(lambda x:x[0])
        newfc.drop(['c1'],inplace=True,axis=1)
        newfc.reset_index(drop=True,inplace=True)
        #print(newfc)
        #exit()
        newfc['c'] = log2(newfc['c'] + 0.05) # * np.min(newfc['c'])
        fc = newfc
        #print("hello")
        #newfc.to_csv("wei.cls",sep="\t")
        #exit()


        rownames = sorted(set(self.usermarkers[ccol].unique()))
        rownum = len(rownames)
        colnames = sorted(set(fc[gcol].unique()))
        colnum = len(colnames)

        #print(fc,rownames,colnames)
        #print(fc.shape)
        return fc,rownames,rownum,colnames,colnum

    def get_cell_matrix(self,exps,ltitle,fid,gcol,ccol,abs_tag):
        """combine cell matrix with weight-matrix"""
        cell_value = None
        colnames = None
        if not self.norefdb:
            cell_value,colnames =self.get_cell_matrix_detail(exps,ltitle,fid,gcol,ccol,False,abs_tag)
        if self.MarkerDB != None:
            if self.norefdb:
                cell_value,colnames =self.get_cell_matrix_detail(exps,ltitle,fid,gcol,ccol,True,abs_tag)
            else:
                cell_value,colnames =self.get_cell_matrix_detail(exps,ltitle,fid,gcol,ccol,False,abs_tag)
                user_value,user_colnames =self.get_cell_matrix_detail(exps,ltitle,fid,gcol,ccol,True,abs_tag)
                #print("C",cell_value)
                #print("U",user_value)
                if cell_value is None:
                    if user_value is None:
                        if colnames is None:
                            return DataFrame(),None
                        else:
                            return DataFrame(),set(colnames)
                    else:
                        cell_value = user_value
                        colnames = user_colnames
                        cell_value = cell_value.join(user_value,how="outer",lsuffix="cm",rsuffix="ur")
                        cell_value[cell_value.isna()] = 0
                        colnames = colnames | user_colnames
                else:
                    if user_value is None:
                        user_value = cell_value
                        user_colnames = colnames
                    cell_value = cell_value.join(user_value,how="outer",lsuffix="cm",rsuffix="ur")
                    cell_value[cell_value.isna()] = 0
                    colnames = colnames | user_colnames

                #else:
                #    cell_value = user_value
                #    colnames = user_colnames
        #database weight-matrix
        wm = [1]
        if self.MarkerDB != None:
            if self.norefdb:
                wm = [1]
            else:
                wm =[0.1,0.9]
        weight_matrix = mat(wm).T

        if colnames is None:
            return DataFrame(),None

        if cell_value is None:
            return DataFrame(),set(colnames)

        #print(cell_value)
        last_value = array(cell_value) * weight_matrix
        result = DataFrame({"Cell Type":cell_value.index,"Z-score":last_value.A1})
        result = result.sort_values(by="Z-score",ascending = False)
        #if self.target == "cancersea":
        #    result['note'] = result['Cell Type'].apply(lambda x: self.snames[x])
        return result,set(colnames)

    def get_cell_matrix_detail(self,exps,ltitle,fid,gcol,ccol,usertag,abs_tag):
        """calculate the cell type scores"""
        fc,rownames,rownum,colnames,colnum = None,None,None,None,None
        #print(self.cmarkers)
        if self.target == "cellmarker":
            markers = self.cmarkers
        elif self.target == "cancersea":
            markers = self.smarkers
        elif self.target == "panglaodb":
            markers = self.pmarkers

        #print(markers.columns)

        if usertag:
            fc,rownames,rownum,colnames,colnum = self.get_user_cell_gene_names(exps,fid,gcol,ccol,"user_marker")
            #print("FC",fc)
            #fc['c'] = 1
        else:
            fc,rownames,rownum,colnames,colnum = self.get_cell_gene_names(exps,markers,fid,gcol,ccol,'marker')
            #print(colnames)
        #print(colnames)

        if not colnames:
            return None,None
        if fc.shape[0] == 0:
            return None,set(colnames)

        exps = exps[exps[fid].isin(colnames)]

        rowdic = dict(zip(rownames,range(rownum)))
        coldic = dict(zip(colnames,range(colnum)))
        fc_cell = fc[ccol].map(lambda x:rowdic[x])
        fc_gene = fc[gcol].map(lambda x:coldic[x])

        newdf = DataFrame({ccol:fc_cell,gcol:fc_gene,"c":fc['c']})
        cell_coo_matrix = coo_matrix((newdf['c'],(newdf[ccol],newdf[gcol])),shape=(rownum,colnum))
        cell_matrix = cell_coo_matrix.toarray()

        #print(newdf)
        #print(rownames)
        #print(colnames)
        #print(cell_matrix)
        

        if self.noprint == False:
            if usertag:
                print("User Cell Num:",rownum)
                print("User Gene Num:",colnum)
                print("User Not Zero:",cell_coo_matrix.count_nonzero())
            else:
                print("Cell Num:",rownum)
                print("Gene Num:",colnum)
                print("Not Zero:",cell_coo_matrix.count_nonzero())
        cell_values = self.get_exp_matrix_loop(exps,ltitle,fid,colnames,rownames,cell_matrix,usertag,abs_tag)
        #print(cell_values)
        return cell_values,set(colnames)


    def read_user_markers(self,colname):
        """usermarker db preparation"""
        if self.MarkerDB != None:
            if not os.path.exists(self.MarkerDB):
                print("User marker database does not exists!",self.MarkerDB)
                sys.exit(0)
            self.usermarkers = read_csv(self.MarkerDB,sep="\t",header=None)
            self.usermarkers.columns=['cellName',colname]
            #self.hgncs_ensem = dict(zip(self.ensem_hgncs.values(),self.ensem_hgncs.keys()))
            if colname == "ensemblID":
                self.usermarkers[colname] = self.usermarkers[colname].map(lambda x:self.hgncs_ensem[x] if x in self.hgncs_ensem else x)
            self.usermarkers['weight'] = 1
            if self.noprint == False:
                print("User cells:", len(self.usermarkers['cellName'].unique()))
                print("User genes:", len(self.usermarkers[colname].unique()))
    
    def save_pickle_module(self, db):
        """save whole database"""
        handler = gzip.open(db, "wb")
        dump(self.gos, handler)
        dump(self.human_gofs, handler)
        dump(self.mouse_gofs, handler)
        dump(self.cmarkers, handler)
        dump(self.smarkers, handler)
        dump(self.snames, handler)
        dump(self.ensem_hgncs, handler)
        dump(self.ensem_mouse, handler)
        
        if 'plus' in db:
            dump(self.pmarkers, handler)
        else:
            dump(self.smarkers, handler)
        
        handler.close()

    def load_pickle_module(self,db):
        """read whole database"""
        if '2023' in db:
            self.year=2023
            if pd.__version__ > "1.5.3":
                print("2023 database build on pandas<2, please downgrade your pandas version!")
                raise ValueError("2023 database build on pandas<2, please downgrade your pandas version!")
        elif '2024' in db:
            self.year=2024
            if pd.__version__ <= "1.5.3":
                print("2024 database build on pandas>2 or higher, please update your pandas version!")
                raise ValueError("2024 database build on pandas>2 or higher, please update your pandas version!")
        else:
            self.year=2023
            print("Your database version is unknown, please check your database version!")

        handler = gzip.open(db,"rb")
        self.gos = load(handler)
        self.human_gofs = load(handler)
        self.mouse_gofs = load(handler)
        self.cmarkers = load(handler)
        self.smarkers = load(handler)
        self.snames = load(handler)
        self.ensem_hgncs = load(handler)
        self.ensem_mouse = load(handler)
        if 'plus' in db:
            self.pmarkers = load(handler)
        else:
            self.pmarkers = self.smarkers.copy()
        self.hgncs_ensem = dict(zip(self.ensem_hgncs.values(),self.ensem_hgncs.keys()))
        fil = []
        #fil = ['Cancer stem cell', 'Cancer cell']
        #print(self.cmarkers)
        #exit()
        self.cmarkers = self.cmarkers[~self.cmarkers['cellName'].isin(fil)]
        if self.cellrange!=None:
            self.cmarkers = self.cmarkers[self.cmarkers['cellName'].str.contains(self.cellrange)]
        #self.smarkers = self.smarkers[~self.smarkers['cellName'].isin(fil)]
        
        #if self.noprint == False:
        if self.year == 2023:
            print("Version V2.1 [2023/06/27]")
        elif self.year == 2024:
            print("Version V2.2 [2024/12/18]")
        else:
            print("Unknown version")
        if 'plus' not in db:
            print("DB load: GO_items:{},Human_GO:{},Mouse_GO:{},\nCellMarkers:{},CancerSEA:{},\nEnsembl_HGNC:{},Ensembl_Mouse:{}".format(
                    len(self.gos),len(self.human_gofs),len(self.mouse_gofs),len(self.cmarkers),len(self.smarkers),len(self.ensem_hgncs),len(self.ensem_mouse))
            )
        else:
            print("DB load: GO_items:{},Human_GO:{},Mouse_GO:{},\nCellMarkers:{},CancerSEA:{},PanglaoDB:{}\nEnsembl_HGNC:{},Ensembl_Mouse:{}".format(
                    len(self.gos),len(self.human_gofs),len(self.mouse_gofs),len(self.cmarkers),len(self.smarkers),len(self.pmarkers),len(self.ensem_hgncs),len(self.ensem_mouse))
            )

    def read_tissues_species(self,tissue="All",species="Human",celltype="normal"):
        """read markers according to certain tissue and certain species"""
        species = species.lower().capitalize()
        ct = celltype.lower().capitalize()
        if tissue != "All":
            if tissue.find(",") > -1:
                temptis = set(tissue.split(","))
                self.cmarkers = self.cmarkers[self.cmarkers['tissueType'].isin(temptis)]
            else:
                self.cmarkers = self.cmarkers[self.cmarkers['tissueType'].isin([tissue])]
        if ct == "Normal":
            self.cmarkers = self.cmarkers[self.cmarkers['cellType']=="Normal cell"]
        elif ct == "Cancer":
            self.cmarkers = self.cmarkers[self.cmarkers['cellType']=="Cancer cell"]
        else:
            print("Illegal celltype. Please use \"[Normal] or [Cancer] instead.")
            exit(0)


        #self.cmarkers = self.cmarkers[self.cmarkers['cellName']!="Mesenchymal stem cell"]
        print("load markers:",len(self.cmarkers))
        self.cmarkers = self.cmarkers[self.cmarkers['speciesType'].isin([species])]
        #print(self.cmarkers)

    def get_list_tissue(self,species):
        """print tissue names"""
        species = species.lower().capitalize()
        cmarkers = self.cmarkers[self.cmarkers['speciesType'].isin([species])]
        names = list(sorted(cmarkers['tissueType'].unique()))
        print("#" * 120)
        print("-" * 120)
        print("{0:s}{1:<10s}{2:>5s}{3:<10d}".format("Species:",species,"Num:",len(names)))
        print("-" * 120)
        for i in range(0,len(names)-2,3):
            if len(names) < i + 1:
                s = "{0:3d}: {1:<40s}".format(i+1,names[i])
            elif len(names) < i + 2:
                s = "{0:3d}: {1:<35s}{2:3d}: {3:<35s}".format(i+1,names[i],i+2,names[i+1])
            else:
                s = "{0:3d}: {1:<35s}{2:3d}: {3:<35s}{4:3d}: {5:<35s}".format(i+1,names[i],i+2,names[i+1],i+3,names[i+2])
            print(s)
        print("#" * 120)




    def run_detail_cmd(self):
        """main command"""
        #self.check_db()
        if not os.path.exists(self.input):
            tempname = "./" + self.input
            if not os.path.exists(tempname):
                print(tempname)
                print("Input file does not exists!",self.input)
                sys.exit(0)
        print(self)
        if self.source.lower() == "cellranger":
            self.load_pickle_module(self.db)
            if self.species == "Mouse":
                self.ensem_hgncs = self.ensem_mouse
                self.human_gofs = self.mouse_gofs
            self.read_tissues_species(self.tissue,self.species,self.celltype)
            if self.Gensymbol:
                self.read_user_markers('Gene ID')
            else:
                self.read_user_markers('ensemblID')
            outs = self.calcu_cellranger_group(self.input,self.Gensymbol)
            return outs
        elif self.source.lower() == "seurat":
            self.load_pickle_module(self.db)
            if self.species == "Mouse":
                self.ensem_hgncs = self.ensem_mouse
                self.human_gofs = self.mouse_gofs
            self.read_tissues_species(self.tissue,self.species,self.celltype)
            if self.Gensymbol:
                self.read_user_markers('gene')
            else:
                self.read_user_markers('ensemblID')
            outs = self.calcu_seurat_group(self.input,self.Gensymbol)
            return outs
        elif self.source.lower() == "scanpy":
            self.load_pickle_module(self.db)
            if self.species == "Mouse":
                self.ensem_hgncs = self.ensem_mouse
                self.human_gofs = self.mouse_gofs
            self.read_tissues_species(self.tissue,self.species,self.celltype)
            if self.Gensymbol:
                self.read_user_markers('gene')
            else:
                self.read_user_markers('ensemblID')
            outs = self.calcu_scanpy_group(self.input,self.Gensymbol)
            return outs
        elif self.source.lower() == "scran":
            self.load_pickle_module(self.db)
            if self.species == "Mouse":
                self.ensem_hgncs = self.ensem_mouse
                self.human_gofs = self.mouse_gofs
            self.read_tissues_species(self.tissue,self.species,self.celltype)
            if self.Gensymbol:
                self.read_user_markers('gene')
            else:
                self.read_user_markers('ensemblID')
            outs = self.calcu_scran_group(self.input,self.Gensymbol)
            return outs
            pass



class Process(object):
    def __init__(self):
        pass
    def get_parser(self):
        desc = """Program: SCSA
  Version: 1.0
  Email  : <yhcao@ibms.pumc.edu.cn>
        """
        #usage = "%(prog)s"
        parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=desc)
                                         #usage=usage)

        parser.add_argument('-i', '--input', required = True, help="Input file for marker annotation(Only CSV format supported).")
        parser.add_argument('-o', '--output', help="Output file for marker annotation.")
        parser.add_argument('-d', '--db', default = "whole.db",help="Database for annotation. (whole.db)")
        parser.add_argument('-s', '--source', default = "cellranger",help="Source of marker genes. (cellranger,[seurat],[scanpy],[scran])")
        parser.add_argument('-c', '--cluster', default = "all",help="Only deal with one cluster of marker genes. (all,[1],[1,2,3],[...])")
        parser.add_argument('-M', '--MarkerDB', help='User-defined marker database in table format with two columns.First column as Cellname, Second refers to Genename.')
        parser.add_argument('-f',"--foldchange",default = 2,help="Fold change threshold for marker filtering. (2.0)")
        parser.add_argument('-p',"--pvalue",default = 0.05,help="P-value threshold for marker filtering. (0.05)")
        parser.add_argument('-w',"--weight",default = 100,help="Weight threshold for marker filtering from cellranger v1.0 results. (100)")
        parser.add_argument('-g',"--species",default = 'Human',help="Species for annotation. Only used for cellmarker database. ('Human',['Mouse'])")
        parser.add_argument('-k',"--tissue",default = 'All',help="Tissue for annotation. Only used for cellmarker database. Multiple tissues should be seperated by commas. Run '-l' option to see all tissues. ('All',['Bone marrow'],['Bone marrow,Brain,Blood'][...])")
        parser.add_argument('-m', '--outfmt', default = "ms-excel", help="Output file format for marker annotation. (ms-excel,[txt])")
        parser.add_argument('-T',"--celltype",default = "normal",help="Cell type for annotation. (normal,[cancer])")
        parser.add_argument('-t', '--target', default = "cellmarker",help="Target to annotation class in Database. (cellmarker,[cancersea,panglaodb])")
        parser.add_argument('-E',"--Gensymbol",action = "store_true",default=False,help="Using gene symbol ID instead of ensembl ID in input file for calculation.")
        parser.add_argument('-N',"--norefdb",action = "store_true",default=False,help="Only using user-defined marker database for annotation.")
        parser.add_argument('-b',"--noprint",action = "store_true",default=False,help="Do not print any detail results.")
        parser.add_argument('-l',"--list_tissue",action = "store_true",default = False,help="List tissue names in database.")
        return parser

    def run_cmd(self,args):
        args.foldchange = float(args.foldchange)
        args.weight = float(args.weight)
        args.pvalue = float(args.pvalue)
        if args.tissue.find(",") > -1:
            tissues = args.tissue
            temptis  = []
            for tis in tissues.split(","):
                temptis.append(str.capitalize(tis))
            args.tissue = ",".join(temptis)
        else:
            args.tissue = str.capitalize(args.tissue)
        args.species = str.capitalize(args.species)
        if args.species == "Mouse" and (args.target == "cancersea" or args.target == "panglaodb"):
            print("Error target database for mouse genome. Cancersea can't used on mouse genomes. Please use cellmarker database instead.")
            exit(0)
        if args.norefdb and not args.MarkerDB:
            print("User-defined marker database must be defined first (-M).")
            exit(0)
        rdbname = Process.check_db(args.db)
        anno = Annotator(args)
        anno.load_pickle_module(rdbname)
        outs = anno.run_detail_cmd()
        print("#Cluster","Type","Celltype","Score","Times")
        for o in outs:
            print(o)
        pass

    def run_cmd_p(self,foldchange,weight,
                 pvalue,tissue,species,
                 target,norefdb,MarkerDB,db,
                 noprint,input,output,source,cluster,fc,
                 outfmt,celltype,Gensymbol,list_tissue,cellrange):
        if tissue.find(",")>-1:
            temptis=[]
            for tis in tissue.split(','):
                temptis.append(str.capitalize(tis))
            tissue=",".join(temptis)
        else:
            tissue=str.capitalize(tissue)
        species=str.capitalize(species)
        if species=="Mouse" and (target=="cancersea" or target=="panglaodb"):
            print("Error target database for mouse genome. Cancersea can't used on mouse genomes. Please use cellmarker database instead.")
            return 0
        if norefdb and not MarkerDB:
            print("User-defined marker database must be defined first (-M).")
            return 0
        rdbname=Process.check_db(db)
        anno = Annotator(foldchange,weight,
                 pvalue,tissue,species,
                 target,norefdb,MarkerDB,db,
                 noprint,input,output,source,cluster,fc,
                 outfmt,celltype,Gensymbol,list_tissue,cellrange)
        anno.load_pickle_module(rdbname)
        outs=anno.run_detail_cmd()
        print("#Cluster","Type","Celltype","Score","Times")
        for o in outs:
            print(o)
        pass
        #anno = Anno



    @staticmethod
    def list_tissue(args):
        rdbname = Process.check_db(args.db)
        anno = Annotator(args)
        anno.load_pickle_module(rdbname)
        anno.get_list_tissue("Human")
        anno.get_list_tissue("Mouse")
        #sys.exit(0)

    @staticmethod
    def check_db(dbname):
        if not os.path.exists(dbname):
            dirname,fname = os.path.split(os.path.realpath(__file__))
            tempname = dirname + "/" + dbname
            if not os.path.exists(tempname):
                print(tempname)
                print("Database does not exists!",dbname)
                sys.exit(0)
        return dbname





import matplotlib.pyplot as plt
import numpy as np
from pandas.api.types import CategoricalDtype, is_numeric_dtype
import seaborn as sns
import scanpy as sc
from scanpy.plotting._anndata import _prepare_dataframe
import pandas as pd
from anndata import AnnData
from ..utils import plotset

pycomplexheatmap_install=False

def check_pycomplexheatmap():
    r"""
    Check if PyComplexHeatmap is installed and available.
    
    Returns:
        None: Prints version information or raises ImportError
    """
    global pycomplexheatmap_install
    try:
        import PyComplexHeatmap as pch
        pycomplexheatmap_install=True
        print('PyComplexHeatmap have been install version:',pch.__version__)
    except ImportError:
        raise ImportError(
            'Please install the tangram: `pip install PyComplexHeatmap`.'
            )
    
def complexheatmap(adata,
                       groupby ='',
                       figsize =(6,10),
                       layer: str = None,
                       use_raw: bool = False,
                       var_names = None,
                       gene_symbols = None,
                       standard_scale:str = None,
                       col_color_bars:dict = None,
                       col_color_labels:dict = None,
                       left_color_bars:dict = None,
                       left_color_labels:dict = None,
                       right_color_bars:dict = None,
                       right_color_labels:dict = None,
                       marker_genes_dict:dict = None,
                       index_name:str = '',
                       value_name:str = '',
                       cmap:str = 'parula',
                       xlabel:str = None,
                       ylabel:str = None,
                       label:str = '',
                       save:bool = False,
                       save_pathway:str = '',
                       legend_gap:int = 7,
                       legend_hpad:int = 0,
                       show:bool = False,
                       left_add_text:bool = False,
                       col_split_gap:int = 1,
                       row_split_gap:int = 1,
                       col_height:int = 4,
                       left_height:int = 4,
                       right_height:int = 4,
                       col_cluster:bool = False,
                       row_cluster:bool = False,
                       row_split = None,
                       col_split = None,
                       legend:bool = True,
                       plot_legend:bool = True,
                       right_fontsize:int = 12,
                        ):
    r"""
    Generate a complex annotated heatmap from single-cell RNA-seq data using PyComplexHeatmap.
    
    Arguments:
        adata: Annotated data object containing expression data
        groupby: Grouping variable for cell categorization ('')
        figsize: Figure dimensions as (width, height) ((6,10))
        layer: Data layer to use for expression values (None)
        use_raw: Whether to use raw expression data (False)
        var_names: List of genes to include in heatmap (None, uses all)
        gene_symbols: Gene symbol column name (None)
        standard_scale: Standardization method - 'obs', 'var', or None (None)
        col_color_bars: Color mapping dictionary for column annotations (None)
        col_color_labels: Color mapping for column labels (None)
        left_color_bars: Color mapping for left annotations (None)
        left_color_labels: Color mapping for left labels (None)
        right_color_bars: Color mapping for right annotations (None)
        right_color_labels: Color mapping for right labels (None)
        marker_genes_dict: Dictionary mapping cell types to marker genes (None)
        index_name: Name for index column in melted data ('')
        value_name: Name for value column in melted data ('')
        cmap: Colormap for heatmap values ('parula')
        xlabel: X-axis label (None)
        ylabel: Y-axis label (None)
        label: Plot label ('')
        save: Whether to save the plot (False)
        save_pathway: File path for saving ('')
        legend_gap: Gap between legend items (7)
        legend_hpad: Horizontal padding for legend (0)
        show: Whether to display the plot (False)
        left_add_text: Whether to add text to left annotations (False)
        col_split_gap: Gap between column splits (1)
        row_split_gap: Gap between row splits (1)
        col_height: Height of column annotations (4)
        left_height: Height of left annotations (4)
        right_height: Height of right annotations (4)
        col_cluster: Whether to cluster columns (False)
        row_cluster: Whether to cluster rows (False)
        row_split: Row splitting variable (None)
        col_split: Column splitting variable (None)
        legend: Whether to show legend (True)
        plot_legend: Whether to plot legend (True)
        right_fontsize: Font size for right annotations (12)
        
    Returns:
        cm: PyComplexHeatmap ClusterMapPlotter object
    """
    check_pycomplexheatmap()
    global pycomplexheatmap_install
    if pycomplexheatmap_install==True:
        global_imports("PyComplexHeatmap","pch")
    
    #sns.set_style('white')
    plotset()
    if layer is not None:
        use_raw = False
    if use_raw == True:
        adata = adata.raw.to_adata()
        use_raw = False
        
    if var_names == None:
        var_names = adata.var_names
    if isinstance(var_names, str):
        var_names = [var_names]
    groupby_copy = groupby
    
    groupby_index = None
    if groupby is not None:
        if isinstance(groupby, str):
            # if not a list, turn into a list
            groupby = [groupby]
        for group in groupby:
            if group not in list(adata.obs_keys()) + [adata.obs.index.name]:
                if adata.obs.index.name is not None:
                    msg = f' or index name "{adata.obs.index.name}"'
                else:
                    msg = ""
                raise ValueError("groupby has to be a valid observation."
                                 f"Given {group}, is not in observations: {adata.obs_keys()}" + msg)
            if group in adata.obs.keys() and group == adata.obs.index.name:
                raise ValueError(f"Given group {group} is both and index and a column level, "
                                 "which is ambiguous.")
            if group == adata.obs.index.name:
                groupby_index = group
    if groupby_index is not None:
        # obs_tidy contains adata.obs.index
        # and does not need to be given
        groupby = groupby.copy()  # copy to not modify user passed parameter
        groupby.remove(groupby_index)

    if col_color_bars == None:
        print('Error, please input col_color before run this function.')
    if col_color_labels == None:
        print('Error, please input col_color before run this function.')

    if marker_genes_dict == None:
        print('Error, please input marker_genes_dict before run this function.')
    
    keys = list(groupby) + list(np.unique(var_names))
    obs_tidy = sc.get.obs_df(adata, keys=keys, layer=layer, use_raw=use_raw, gene_symbols=gene_symbols)
    assert np.all(np.array(keys) == np.array(obs_tidy.columns))

    if groupby_index is not None:
        # reset index to treat all columns the same way.
        obs_tidy.reset_index(inplace=True)
        groupby.append(groupby_index)

    if groupby is None:
        categorical = pd.Series(np.repeat("", len(obs_tidy))).astype("category")
    elif len(groupby) == 1 and is_numeric_dtype(obs_tidy[groupby[0]]):
        # if the groupby column is not categorical, turn it into one
        # by subdividing into  `num_categories` categories
        categorical = pd.cut(obs_tidy[groupby[0]], num_categories)
    elif len(groupby) == 1:
        categorical = obs_tidy[groupby[0]].astype("category")
        categorical.name = groupby[0]
    else:
        # join the groupby values  using "_" to make a new 'category'
        categorical = obs_tidy[groupby].apply("_".join, axis=1).astype("category")
        categorical.name = "_".join(groupby)

        # preserve category order
        from itertools import product

        order = {
            "_".join(k): idx
            for idx, k in enumerate(
                product(*(obs_tidy[g].cat.categories for g in groupby))
            )
        }
        categorical = categorical.cat.reorder_categories(
            sorted(categorical.cat.categories, key=lambda x: order[x])
        )
    obs_tidy = obs_tidy[var_names].set_index(categorical)
    categories = obs_tidy.index.categories
    obs_tidy = obs_tidy.groupby(groupby).mean()


    
    if standard_scale == "obs":
        obs_tidy = obs_tidy.sub(obs_tidy.min(1), axis=0)
        obs_tidy = obs_tidy.div(obs_tidy.max(1), axis=0).fillna(0)
    elif standard_scale == "var":
        obs_tidy -= obs_tidy.min(0)
        obs_tidy = (obs_tidy / obs_tidy.max(0)).fillna(0)
    elif standard_scale is None:
        pass


    if right_color_bars==None:
        # colorbar of gene
        gene_color_dict = {}
        for cell_type, genes in marker_genes_dict.items():
            cell_type_color = [color for color, category in zip(adata.uns[groupby_copy+'_colors'], adata.obs[groupby_copy].cat.categories) if category == cell_type][0]
            for gene in genes:
                gene_color_dict[gene] = cell_type_color
        right_color_bars = gene_color_dict

    if right_color_labels==None:
        right_color_labels = right_color_bars


    # col
    df_col = obs_tidy.copy()
    df_col[groupby_copy] = df_col.index
    col_ha = pch.HeatmapAnnotation(label=pch.anno_label(df_col[groupby_copy],merge=True,rotation=90,extend=True,
                                            colors=col_color_bars,adjust_color=True,luminance=0.75,
                                            relpos=(0.5,0)), #fontsize=10
                           Celltype=pch.anno_simple(df_col[groupby_copy],colors=col_color_labels,height=col_height), #legend_kws={'fontsize':4}
                           verbose=1,axis=1,plot=False)

    
    # dict to Dataframe
    marker_genes_df = pd.DataFrame.from_dict(marker_genes_dict, orient='index')
    # Dataframe transpose
    marker_genes_df = marker_genes_df.transpose()
    melted_df = marker_genes_df.melt(var_name=index_name, value_name=value_name).dropna()
    melted_df.index = melted_df.loc[:,value_name]
    df_row = melted_df
    del melted_df

    if left_color_labels == None:
        left_ha = pch.HeatmapAnnotation(
                           Marker_Gene=pch.anno_simple(df_row[index_name],legend=True,
                                             colors=left_color_bars,add_text=left_add_text,height=left_height),
                           verbose=1,axis=0,plot_legend=False,plot=False)
    else:   
        left_ha = pch.HeatmapAnnotation(
                           label=pch.anno_label(df_row[index_name],merge=True,extend=False,
                                            colors=left_color_labels,adjust_color=True,luminance=0.75,
                                              relpos=(1,0.5)),
                           Marker_Gene=pch.anno_simple(df_row[index_name],legend=True,
                                             colors=left_color_bars,add_text=left_add_text,height=left_height),
                           verbose=1,axis=0,plot_legend=False,plot=False)

    if right_color_labels == None:
        right_ha = pch.HeatmapAnnotation(
                           Group=pch.anno_simple(df_row[index_name],legend=True,
                                             colors=right_color_bars,height=right_height),
                           verbose=1,axis=0,plot_legend=False,label_kws=dict(visible=False),plot=False,)
    else:
        right_ha = pch.HeatmapAnnotation(
                           Group=pch.anno_simple(df_row[index_name],legend=True,
                                             colors=right_color_bars,height=right_height),
                           label=pch.anno_label(df_row[value_name],merge=True,extend=True,
                                            colors=right_color_labels,adjust_color=True,luminance=0.75,
                                            relpos=(0,0.5),fontsize=right_fontsize), #fontsize=10
                           verbose=1,axis=0,plot_legend=False,label_kws=dict(visible=False),plot=False)        



    if row_split!=None:
        row_split_copy = df_row.loc[:,index_name]
    else:
        row_split_copy = row_split
    if col_split!=None:
        col_split_copy = df_col.loc[:,index_name]
    else:
        col_split_copy = col_split


    plt.figure(figsize=figsize)
    obs_copy = obs_tidy.copy().loc[df_col.index.tolist(),df_row.index.tolist()]
    cm = pch.ClusterMapPlotter(data=obs_copy.T,
                       top_annotation=col_ha, 
                       left_annotation=left_ha,
                       right_annotation=right_ha,
                       row_cluster=row_cluster,col_cluster=col_cluster,
                       label=label, row_dendrogram=False,legend_gap=legend_gap,
                       row_split=row_split_copy,col_split=col_split_copy,
                       col_split_gap=col_split_gap,
                       row_split_gap=row_split_gap,
                       row_split_order=list(marker_genes_dict.keys()),
                       # col_split_order=df_row.Group.unique().tolist(),
                       cmap=cmap,rasterized=True,
                       xlabel=xlabel, legend_hpad=legend_hpad,
                       ylabel=ylabel,
                       xlabel_kws=dict(color='black', fontsize=14, labelpad=0),
                       legend=legend,
                       plot_legend=plot_legend,
                       # ylabel_kws=dict(color='black', fontsize=14, labelpad=0),
                          )
    
    #plt.savefig("Loyfer2023_heatmap.pdf",bbox_inches='tight')
    if save ==True:
        plt.savefig(save_pathway,bbox_inches='tight',dpi=300)
    if show ==True:
        plt.show()
    return cm

def marker_heatmap(
    adata: AnnData,
    marker_genes_dict: dict = None,
    groupby: str = None,
    color_map: str = "RdBu_r",
    use_raw: bool = True,
    standard_scale: str = "var",
    expression_cutoff: float = 0.0,
    bbox_to_anchor: tuple = (5, -0.5),
    figsize: tuple = (8,4),
    spines: bool = False,
    fontsize: int = 12,
    show_rownames: bool = True,
    show_colnames: bool = True,
    save_path: str = None,
    ax=None,
):
    r"""
    Create a dot plot heatmap showing marker gene expression across cell types.
    
    Arguments:
        adata: Annotated data object with expression data
        marker_genes_dict: Dictionary mapping cell types to marker genes (None)
        groupby: Column name for cell type grouping (None)
        color_map: Colormap for expression values ('RdBu_r')
        use_raw: Whether to use raw expression data (True)
        standard_scale: Expression standardization method - 'var', 'group', or None ('var')
        expression_cutoff: Minimum expression threshold for dot display (0.0)
        bbox_to_anchor: Legend position as (x, y) coordinates ((5, -0.5))
        figsize: Figure dimensions as (width, height) ((8,4))
        spines: Whether to show plot spines (False)
        fontsize: Font size for labels and text (12)
        show_rownames: Whether to display row names (True)
        show_colnames: Whether to display column names (True)
        save_path: File path for saving plot (None)
        ax: Existing matplotlib axes object (None)
        
    Returns:
        fig: matplotlib.figure.Figure object
        ax: matplotlib.axes.Axes object
    """
    
    # input check
    if marker_genes_dict is None:
        print("Please provide a dictionary containing the marker genes for each cell type.")
        return
    if groupby is None:
        print("Please provide a key in adata.obs for grouping the cells.")  
        return

    # pycomplexheatmap version check
    try:
        import PyComplexHeatmap as pch
        from PyComplexHeatmap import DotClustermapPlotter,HeatmapAnnotation,anno_simple,anno_label,AnnotationBase
        print('PyComplexHeatmap have been install version:',pch.__version__)
        if pch.__version__ < '1.7.5':
            raise ImportError(
            'Please install PyComplexHeatmap with version > 1.7.5: `pip install PyComplexHeatmap`.'
            )
    except ImportError:
        raise ImportError(
            'Please install PyComplexHeatmap with version > 1.7.5: `pip install PyComplexHeatmap`.'
            )

     # Determine the color palette for different categories based on annotation data.
    if f"{groupby}_colors" in adata.uns:
        type_color_all = dict(zip(adata.obs[groupby].cat.categories,adata.uns[f"{groupby}_colors"]))
    else:
        if '{}_colors'.format(groupby) in adata.uns:
            type_color_all=dict(zip(adata.obs[groupby].cat.categories,adata.uns['{}_colors'.format(groupby)]))
        else:
            if len(adata.obs[groupby].cat.categories)>28:
                type_color_all=dict(zip(adata.obs[groupby].cat.categories,sc.pl.palettes.default_102))
            else:
                type_color_all=dict(zip(adata.obs[groupby].cat.categories,sc.pl.palettes.zeileis_28))

    # Prepare lists to hold gene group labels and positions.
    var_group_labels = []
    _var_names = []
    var_group_positions = []
    start = 0
    for label, vars_list in marker_genes_dict.items():
        if isinstance(vars_list, str):
            vars_list = [vars_list]
        _var_names.extend(list(vars_list))
        var_group_labels.append(label)
        var_group_positions.append((start, start + len(vars_list) - 1))

    # Prepare data for plotting using Scanpy's internal function.
    categories, obs_tidy = _prepare_dataframe(
            adata,
            _var_names,
            groupby=groupby,
            use_raw=use_raw,
            log=False,
            num_categories=7,
            layer=None,
            gene_symbols=None,
        )

    # determine the dot size and calculate the mean expression and fraction of cells.
    obs_bool = obs_tidy > expression_cutoff
    dot_size_df = (
                    obs_bool.groupby(level=0, observed=True).sum()
                    / obs_bool.groupby(level=0, observed=True).count()
                )

    # Standardize the expression values
    dot_color_df = obs_tidy.groupby(level=0, observed=True).mean()
    if standard_scale == "group":
        dot_color_df = dot_color_df.sub(dot_color_df.min(1), axis=0)
        dot_color_df = dot_color_df.div(dot_color_df.max(1), axis=0).fillna(0)
    elif standard_scale == "var":
        dot_color_df -= dot_color_df.min(0)
        dot_color_df = (dot_color_df / dot_color_df.max(0)).fillna(0)
    elif standard_scale is None:
        pass

    # Data preparation for pycomplexheatmap
    Gene_list = []
    for celltype in marker_genes_dict.keys():
        for gene in marker_genes_dict[celltype]:
            Gene_list.append(gene)

    # Prepare data for complex heatmap plotting.
    df_row=dot_color_df.index.to_frame()
    df_row['Celltype']=dot_color_df.index
    df_row.set_index('Celltype',inplace=True)
    df_row.columns = ['Celltype_name']
    df_row = df_row.loc[list(marker_genes_dict.keys()),:]

    df_col = pd.DataFrame()
    for celltype in marker_genes_dict.keys():
        df_col_tmp=pd.DataFrame(index = marker_genes_dict[celltype])
        df_col_tmp['Gene']=marker_genes_dict[celltype]
        df_col_tmp['Celltype_name'] = celltype
        df_col = pd.concat([df_col,df_col_tmp])
    df_col.columns = ['Gene_name','Celltype_name']
    df_col = df_col.loc[Gene_list,:]

    # Create a melted DataFrame for color and size data.
    color_df = pd.melt(dot_color_df.reset_index(), id_vars=groupby, var_name='gene', value_name='Mean\nexpression\nin group')
    color_df[groupby] = color_df[groupby].astype(str)
    color_df.index = color_df[groupby]+'_'+color_df['gene']
    size_df = pd.melt(dot_size_df.reset_index(), id_vars=groupby, var_name='gene', value_name='Fraction\nof cells\nin group')
    size_df[groupby] = size_df[groupby].astype(str)
    size_df.index = size_df[groupby]+'_'+size_df['gene']
    color_df['Fraction\nof cells\nin group'] = size_df.loc[color_df.index.tolist(),'Fraction\nof cells\nin group']

    Gene_color = []
    for celltype in df_row.Celltype_name:
        for gene in marker_genes_dict[celltype]:
            Gene_color.append(type_color_all[celltype])

    # plot the complex heatmap
    if ax==None:
        fig, ax = plt.subplots(figsize=figsize) 
    else:
        ax=ax

    row_ha = HeatmapAnnotation(
                TARGET=anno_simple(
                    df_row.Celltype_name,
                    colors=[type_color_all[i] for i in df_row.Celltype_name],
                    add_text=False,
                    text_kws={'color': 'black', 'rotation': 0,'fontsize':fontsize},
                    legend=False  # 设置为 True 以显示行的图例
                ),
                legend_gap=7,
                axis=0,
                verbose=0,
                #label_side='left',
                label_kws={'rotation': 90, 'horizontalalignment': 'right','fontsize':0},
            )

    col_ha = HeatmapAnnotation(
                TARGET=anno_simple(
                    df_col.Gene_name,
                    colors=Gene_color,
                    add_text=False,
                    text_kws={'color': 'black', 'rotation': 0,'fontsize':fontsize},
                    legend=False  # 设置为 True 以显示行的图例
                ),
                verbose=0,
                label_kws={'horizontalalignment': 'right','fontsize':0},
                legend_kws={'ncols': 1},  # 调整图例的列数为1
                legend=False,
                legend_hpad=7,
                legend_vpad=5,
                axis=1,
            )

    cm = DotClustermapPlotter(color_df,y=groupby,x='gene',value='Mean\nexpression\nin group',
                      c='Mean\nexpression\nin group',s='Fraction\nof cells\nin group',cmap=color_map,
                      vmin=0,
                      #hue=groupby,
                      top_annotation=col_ha,left_annotation=row_ha,
                      row_dendrogram=False,col_dendrogram=False,
                      col_split_order=list(df_col.Celltype_name.unique()),
                      col_split=df_col.Celltype_name,col_split_gap=1,
                      xticklabels_kws={'labelsize':fontsize},
                      yticklabels_kws={'labelsize':fontsize},
                      dot_legend_kws={'fontsize':fontsize,
                                      'title_fontsize':fontsize},
                      color_legend_kws={'fontsize':fontsize},
                  #    row_split=df_row.Celltype_name,row_split_gap=1,
                      x_order=df_col.Gene_name.unique(),y_order=df_col.Celltype_name.unique(),
                      row_cluster=False,col_cluster=False,
                      show_rownames=show_rownames,show_colnames=show_colnames,
                      col_names_side='left',spines=spines,grid='minor',
                      legend=True,)

    # Adjust grid settings
    cm.ax_heatmap.grid(which='minor', color='gray', linestyle='--', alpha=0.5)
    cm.ax_heatmap.grid(which='major', color='black', linestyle='-', linewidth=0.5)
    cm.cmap_legend_kws={'ncols': 1}
    plt.grid(False)
    plt.tight_layout()  # 调整布局以适应所有组件

    for ax1 in plt.gcf().axes:
        ax1.grid(False)

    # legend plot
    handles = [plt.Line2D([0], [0], color=type_color_all[cell], lw=4) for cell in type_color_all.keys()]
    labels = type_color_all.keys()
    # Add a legend to the right of the existing image
    legend_kws={'fontsize':fontsize,'bbox_to_anchor':bbox_to_anchor,'loc':'center left',}
    plt.legend(handles, labels, 
        borderaxespad=1, handletextpad=0.5, labelspacing=0.2,**legend_kws)

    if save_path is None:
        pass
    else:
        plt.savefig(save_path, dpi=300, bbox_inches='tight')
        
    plt.tight_layout()
    #plt.show()

    return fig,ax


def global_imports(modulename,shortname = None, asfunction = False):
    r"""
    Import a module into the global namespace with optional short name.
    
    Arguments:
        modulename: Name of the module to import
        shortname: Short name for the module (None, uses modulename)
        asfunction: Whether to import as function (False)
        
    Returns:
        None: Imports module into global namespace
    """
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)


from typing import Optional

import pandas as pd
import numpy as np
import networkx as nx
import scanpy as sc
from sklearn.linear_model import LinearRegression as lr

from matplotlib import pyplot as plt
#from matplotlib_venn import venn3
import matplotlib.ticker as ticker
from matplotlib.pyplot import rc_context
import seaborn as sns
# import aucell and ctxcore
from ._aucell import aucell

from .driver_regulators import driver_regulators

venn_install = False
ctxcore_install = False

def matplotlib_venn():
    global venn_install
    try:
        import matplotlib_venn as venn3
        venn_install=True
    except ImportError:
        raise ImportError(
            'Please install the matplotlib_venn: `pip install matplotlib_venn`.'
        )
    
def check_ctxcore():
    global ctxcore_install
    try:
        from ctxcore.genesig import GeneSignature
        ctxcore_install = True
    except ImportError:
        raise ImportError(
            'Please install the ctxcore: `pip install ctxcore`.'
        )
    
class CefconResults:

    def __init__(self,
                 adata: sc.AnnData,
                 network: nx.DiGraph,
                 gene_embedding: pd.DataFrame):
        self.name = adata.uns['name']
        self.network = network
        self.gene_embedding = gene_embedding

        genes = list(network.nodes())
        self.expression_data = adata[:, genes].to_df()
        self.TFs = adata[:, genes].var_names[adata[:, genes].var['is_TF'] == 1]
        if 'node_score_auxiliary' in adata.var:
            self.DEgenes = set(adata.var_names[adata.var['node_score_auxiliary'] > 0]).intersection(set(genes))
        else:
            self.DEgenes = set(genes)

        self.n_cells = adata.n_obs
        self.n_genes = len(genes)
        self.n_edges = network.number_of_edges()
        self.influence_score = None
        self.driver_regulator = None
        self.gene_cluster = None
        self.RGMs_AUCell_dict = None
        self._adata_gene = None
        self._out_critical_genes = None
        self._in_critical_genes = None

    def __repr__(self) -> str:
        descr = f"CefconResults object with n_cells * n_genes = {self.n_cells} * {self.n_genes}, n_edges = {self.n_edges}"
        descr += f"\n    name: {self.name}" \
                 f"\n    expression_data: yes" \
                 f"\n    network: {self.network}" \
                 f"\n    gene_embedding: # dimension = {self.gene_embedding.shape[1] if self.gene_embedding is not None else 'None'}" \
                 f"\n    influence_score: {'None' if self.influence_score is None else 'yes'}" \
                 f"\n    driver_regulator: {'None' if self.driver_regulator is None else 'yes'}" \
                 f"\n    gene_cluster: {'None' if self.gene_cluster is None else 'yes'}" \
                 f"\n    RGMs_AUCell_dict: {'None' if self.RGMs_AUCell_dict is None else 'yes'}"
        return descr

    def gene_influence_score(self):
        """
        Obtain gene influence score
        """
        influence_score = pd.DataFrame(np.zeros((len(self.network.nodes), 2)),
                                       index=sorted(self.network.nodes),
                                       columns=['out', 'in'])
        for i, v in enumerate(['in', 'out']):
            # The out-degree type of influence is obtained from the incoming network;
            # The in-degree type of influence is obtained from the outgoing network.
            gene_att_score = np.sum(nx.to_numpy_array(self.network,
                                                      nodelist=sorted(self.network.nodes),
                                                      dtype='float32',
                                                      weight='weights_{}'.format(v)),  # {}_weights
                                    axis=1 - i)
            influence_score.iloc[:, i] = np.log1p(gene_att_score).flatten().tolist()

        lam = 0.8
        influence_score['influence_score'] = lam * influence_score.loc[:, 'out'] + \
                                             (1 - lam) * influence_score.loc[:, 'in']
        influence_score.rename(columns={'out': 'score_out', 'in': 'score_in'}, inplace=True)
        influence_score = influence_score.sort_values(by='influence_score', ascending=False)
        self.influence_score = influence_score

    def driver_regulators(self, topK: int = 100, return_value: bool = False, output_file: Optional[str] = None, solver = 'GUROBI'):
        """
        Obtain driver regulators from the constructed cell-lineage-specific GRN
        """
        if self.influence_score is None:
            self.gene_influence_score()

        self.driver_regulator, self._out_critical_genes, self._in_critical_genes = driver_regulators(self.network,
                                                                                                     self.influence_score,
                                                                                                     topK=topK,
                                                                                                     driver_union=True,
                                                                                                     solver=solver)
        self.driver_regulator['is_TF'] = self.driver_regulator.index.isin(self.TFs)

        if isinstance(output_file, str):
            self.driver_regulator.to_csv(output_file)
        if return_value:
            return self.driver_regulator

    def RGM_activity(self, num_workers: int = 8, return_value: bool = False):
        """
        Select RGMs of driver regulators and calculate their activities in each cell.
        Activity score is calculated based on AUCell, which is from the `pyscenic` package.
        """
        print('[3] - Identifying regulon-like gene modules...')

        if self.driver_regulator is None:
            raise ValueError(
                f'Did not find the result of driver regulators. Run `cefcon.NetModel.driver_regulators` first.'
            )


        check_ctxcore()
        global ctxcore_install
        if ctxcore_install==True:
            global GeneSignature
            from ctxcore.genesig import GeneSignature

        network = self.network
        DEgenes = self.DEgenes
        drivers = set(self.driver_regulator[self.driver_regulator['is_driver_regulator']].index)

        auc_mtx_out, auc_mtx_in, auc_mtx = None, None, None
        out_hub_nodes = self._out_critical_genes.intersection(drivers)
        out_RGMs = {i + '_out({})'.format(len(list(set(network.successors(i)).intersection(DEgenes)))):
                        list(set(network.successors(i)).intersection(DEgenes)) for i in out_hub_nodes
                    if len(set(network.successors(i)).intersection(DEgenes)) >= 10}
        in_hub_nodes = self._in_critical_genes.intersection(drivers)
        in_RGMs = {i + '_in({})'.format(len(list(set(network.predecessors(i)).intersection(DEgenes)))):
                       list(set(network.predecessors(i)).intersection(DEgenes)) for i in in_hub_nodes
                   if len(set(network.predecessors(i)).intersection(DEgenes)) >= 10}

        # n_cells x n_genes
        out_RGMs = [GeneSignature(name=k, gene2weight=v) for k, v in out_RGMs.items()]
        in_RGMs = [GeneSignature(name=k, gene2weight=v) for k, v in in_RGMs.items()]
        if len(out_RGMs) > 0:
            auc_mtx_out = aucell(self.expression_data, out_RGMs, num_workers=num_workers, auc_threshold=0.25,
                                 normalize=False)
            # Generate a Z-score for each RGM to enable comparison between RGMs
            auc_mtx_out_Z = pd.DataFrame(index=auc_mtx_out.index, columns=list(auc_mtx_out.columns), dtype=float)
            for col in list(auc_mtx_out.columns):
                auc_mtx_out_Z[col] = (auc_mtx_out[col] - auc_mtx_out[col].mean()) / auc_mtx_out[col].std(ddof=0)

        if len(in_RGMs) > 0:
            auc_mtx_in = aucell(self.expression_data, in_RGMs, num_workers=num_workers, auc_threshold=0.25,
                                normalize=False)
            # Generate a Z-score for each RGM to enable comparison between RGMs
            auc_mtx_in_Z = pd.DataFrame(index=auc_mtx_in.index, columns=list(auc_mtx_out.columns), dtype=float)
            for col in list(auc_mtx_in.columns):
                auc_mtx_in_Z[col] = (auc_mtx_in[col] - auc_mtx_in[col].mean()) / auc_mtx_in[col].std(ddof=0)

        if (len(out_RGMs) > 0) & (len(in_RGMs) > 0):
            auc_mtx = pd.merge(auc_mtx_out, auc_mtx_in, how='inner', left_index=True, right_index=True)
            # Generate a Z-score for each RGM to enable comparison between RGMs
            auc_mtx_Z = pd.DataFrame(index=auc_mtx.index, columns=list(auc_mtx.columns), dtype=float)
            for col in list(auc_mtx.columns):
                auc_mtx_Z[col] = (auc_mtx[col] - auc_mtx[col].mean()) / auc_mtx[col].std(ddof=0)

        RGMs = out_RGMs + in_RGMs
        RGMs_AUCell_dict = {'RGMs': RGMs, 'aucell': auc_mtx,
                            'aucell_out': auc_mtx_out, 'aucell_in': auc_mtx_in}
        self.RGMs_AUCell_dict = RGMs_AUCell_dict
        print('Done!')

        if return_value:
            return RGMs_AUCell_dict

    def plot_gene_embedding_with_clustering(self,
                                            n_neighbors: int = 30,
                                            resolution: float = 1,
                                            return_value: bool = False):
        """
        Cluster genes with CEFCON derived embeddings by using the Leiden algorithm and visualize in 2-D UMAP space.
        """
        if self._adata_gene is None:
            self._adata_gene = sc.AnnData(X=self.gene_embedding)
            sc.pp.neighbors(self._adata_gene, n_neighbors=n_neighbors, use_rep='X')
            # Higher resolutions lead to more communities, while lower resolutions lead to fewer communities.
            sc.tl.leiden(self._adata_gene, resolution=resolution)
            sc.tl.umap(self._adata_gene, n_components=2, min_dist=0.3)
            pos = self._adata_gene.obsm['X_umap']
            self.gene_cluster = self._adata_gene.obs['leiden']

        with rc_context({'figure.figsize': (4, 4)}):
            sc.pl.umap(self._adata_gene, color=['leiden'], legend_loc='on data',
                       legend_fontsize=8, legend_fontoutline=2, title='Leiden clustering using CEFCON derived gene embeddings')

        if return_value:
            return self.gene_cluster

    def plot_influence_score(self, topK: int = 20):
        """
        Plot the gene influence score of top-k driver regulators.
        """
        if self.driver_regulator is None:
            raise ValueError(
                f'Did not find the result of driver regulators. Run `cefcon.NetModel.driver_regulators` first.'
            )
        data_for_plot = self.driver_regulator[self.driver_regulator['is_driver_regulator']]
        data_for_plot = data_for_plot[0:topK]

        plt.figure(figsize=(1.5, topK * 0.15))
        #sns.set_theme(style='ticks', font_scale=0.5)

        ax = sns.barplot(x='influence_score', y=data_for_plot.index, data=data_for_plot, orient='h',
                         palette=sns.color_palette(f"ch:start=.5,rot=-.5,reverse=1,dark=0.4", n_colors=topK))
        ax.set_title(self.name)
        ax.set_xlabel('Influence score')
        ax.set_ylabel('Driver regulators')
        ax.xaxis.set_minor_locator(ticker.NullLocator())
        ax.yaxis.set_minor_locator(ticker.NullLocator())
        sns.despine()

        plt.show()

    def plot_network(self, nodes: Optional[list] = None):
        """
        Plot network graph with circos layout.
        """
        try:
            import nxviz as nv
            from nxviz import annotate
        except ImportError:
            raise ImportError("install nxviz via `pip install nxviz`")

        if nodes is not None:
            sub_net = self.network.subgraph(nodes)
        else:
            sub_net = self.network

        nx.set_node_attributes(sub_net, self.driver_regulator['influence_score'].to_dict(), 'influence_score')
        nx.set_node_attributes(sub_net, {i: i for i in sub_net.nodes}, 'gene_name')

        #sns.set_theme(font_scale=0.7)
        ax = nv.circos(sub_net,
                       sort_by='influence_score',
                       # node_size_by='influence_score',
                       node_enc_kwargs={
                           'size_scale': 0.6,
                       },
                       edge_alpha_by='weights_combined',
                       edge_enc_kwargs={
                           'lw_scale': 0.5,
                           'alpha_scale': 0.7,
                       },
                       )
        annotate.circos_labels(sub_net, sort_by='influence_score', layout='rotate')

    def plot_driver_genes_Venn(self,figsize=(4,4)):
        """
        Plot Venn diagram of MDS, MFVS and top-ranked regulators.
        """
        if self.driver_regulator is None:
            raise ValueError(
                f'Did not find the result of driver regulators. Run `cefcon.NetModel.driver_regulators` first.'
            )
        else:
            drivers_df = self.driver_regulator

        MFVS_driver_set = set(drivers_df.loc[drivers_df['is_MFVS_driver']].index)
        MDS_driver_set = set(drivers_df.loc[drivers_df['is_MDS_driver']].index)
        top_ranked_genes = set(drivers_df.loc[drivers_df['is_driver_regulator']].index).union(
            (set(drivers_df.index) - MFVS_driver_set.union(MDS_driver_set)))

        f = plt.figure(figsize=figsize)
        #sns.set_theme(font_scale=f.get_dpi() / 100)

        matplotlib_venn()
        global venn_install
        if venn_install==True:
            global_imports_members('matplotlib_venn', members=['venn3'], asfunction=True)
            
        out = venn3(subsets=[MDS_driver_set, MFVS_driver_set, top_ranked_genes],
                    set_labels=('MDS driver genes({})'.format(len(MDS_driver_set)),
                                'MFVS driver genes({})'.format(len(MFVS_driver_set)),
                                'Top ranked genes({})'.format(len(top_ranked_genes))),
                    set_colors=('#0076FF', '#D74715', '#009000'),
                    alpha=0.45)
        for text in out.set_labels:
            if text is not None:
                text.set_fontsize(7)
        for text in out.subset_labels:
            if text is not None:
                text.set_fontsize(8)
        plt.tight_layout()
        # plt.savefig('drivers_Venn_plot.svg', dpi=600, format='svg', bbox_inches='tight')
        plt.show()

    def plot_RGM_activity_heatmap(self, cell_label: Optional[pd.DataFrame] = None,
                                   type: str = 'out',col_cluster = True, bbox_to_anchor=(1.35, 0.90)):
        """
        Plot clustermap of RGM activity matrix.
        If `cell_label` is provided, cells are ordered by cell clusters, else cells are ordered by hierarchical clustering.
        """
        if self.RGMs_AUCell_dict is None:
            raise ValueError(
                f'Did not find the result of RGMs. Run `cefcon.NetModel.RGM_activity` first.'
            )
        assert type in ['out', 'in', 'all']

        if type == 'all':
            auc_mtx = self.RGMs_AUCell_dict['aucell']
        elif type == 'out':
            auc_mtx = self.RGMs_AUCell_dict['aucell_out']
        else:  # in
            auc_mtx = self.RGMs_AUCell_dict['aucell_in']

        # Create a categorical palette to identify the networks
        if cell_label is not None:
            network_lable = cell_label
            network_pal = sns.husl_palette(len(network_lable.unique()), h=.5)
            network_lut = dict(zip(map(str, network_lable.unique()), network_pal))
            network_colors = pd.Series(list(network_lable), index=auc_mtx.index).map(network_lut)
            col_cluster = col_cluster
        else:
            network_colors = None
            col_cluster = col_cluster

        # plot clustermap (n_cell * n_gene)
        f = plt.figure()
        #sns.set_theme(font_scale=f.get_dpi()/150)
        g = sns.clustermap(auc_mtx.T, method='ward', square=False, linecolor='black',
                           z_score=0, vmin=-2.5, vmax=2.5,
                           col_cluster=col_cluster, col_colors=network_colors, cmap="RdBu_r",
                           figsize=(4.5, 0.15 * auc_mtx.shape[1]),
                           xticklabels=False, yticklabels=True, dendrogram_ratio=0.12,
                           cbar_pos=(0.75, 0.92, 0.2, 0.02),
                           cbar_kws={'orientation': 'horizontal'})
        g.cax.set_visible(True)
        g.ax_heatmap.set_ylabel('Regulon-like gene modules')
        g.ax_heatmap.set_xlabel('Cells')
        g.ax_heatmap.yaxis.set_minor_locator(ticker.NullLocator())
        if cell_label is not None:
            for label in network_lable.unique():
                g.ax_col_dendrogram.bar(0, 0, color=network_lut[label], label=label, linewidth=0)
            g.ax_col_dendrogram.legend(title='Cell types', loc="upper left", ncol=1,
                                       bbox_to_anchor=bbox_to_anchor, facecolor='white')
        g.ax_row_dendrogram.set_visible(False)
        plt.show()

    def plot_network_degree_distribution(self):
        """
        Plot degree distribution of the predicted network
        """
        network = self.network.copy()
        network.remove_edges_from(nx.selfloop_edges(network))
        network = network.subgraph(max(nx.weakly_connected_components(network), key=len)).copy()

        # Slope of degree distribution
        degree_sequence = pd.DataFrame(np.array(network.degree))
        degree_sequence.columns = ["ind", "degree"]
        degree_sequence['degree'] = degree_sequence['degree'].astype(int)
        degree_sequence = degree_sequence.loc[degree_sequence['degree'] != 0, :]
        degree_sequence = degree_sequence.set_index("ind")
        dist = degree_sequence.degree.value_counts() / degree_sequence.degree.value_counts().sum()
        dist.index = dist.index.astype(int)

        x = np.log(dist.index.values).reshape([-1, 1])
        y = np.log(dist.values).reshape([-1, 1])

        model = lr()
        model.fit(x, y)

        # Figure
        fig, ax = plt.subplots(1, 1, figsize=(4, 3))
        #sns.set_theme(style="white", font_scale=fig.get_dpi() / 120)

        x_ = np.array([-1, 5]).reshape([-1, 1])
        y_ = model.predict(x_)

        ax.set_title("degree distribution (log scale)")
        ax.plot(x_.flatten(), y_.flatten(), c="black", alpha=0.5)

        ax.scatter(x.flatten(), y.flatten(), c="black")
        ax.text(0.45, 0.95,
                f"slope: {model.coef_[0][0] :.4g}, " + r"$R^2$: " + f"{model.score(x, y) :.4g}\n" +
                f"num_of_genes: {self.network.number_of_nodes()}\n" +
                f"num_of_edges: {self.network.number_of_edges()}\n" +
                f"clustering_coefficient: {nx.average_clustering(self.network) :.4g}\n",
                transform=ax.transAxes,
                verticalalignment='top',
                horizontalalignment='left',
                fontsize=8.5)
        ax.set_ylim([y.min() - 0.2, y.max() + 0.2])
        ax.set_xlim([-0.2, x.max() + 0.2])
        ax.set_xlabel("log k")
        ax.set_ylabel("log P(k)")
        ax.grid(None)

        plt.tight_layout()
        plt.show()


def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)

def global_imports_members(modulename, members=None, asfunction=False):
    if members is None:
        members = [modulename]  # Default to importing the entire module

    imported_module = __import__(modulename, fromlist=members)

    if asfunction:
        for member in members:
            globals()[member] = getattr(imported_module, member)
    else:
        globals()[modulename] = imported_module



"""
Module to define general nodes in a Bayesian network

All nodes have two main attributes:
- dim: dimensionality of the node
- markov_blanket: the markov blanket of the node

"""

import scipy as s
import numpy as np

from ... import config


class Node(object):
    """General class for a node in a Bayesian network

    PARAMETERS
    ----------
    dim: tuple
        Dimensionality of the node
    """
    def __init__(self, dim):
        self.dim = dim

    def addMarkovBlanket(self, **kwargs):
        """Method to define the Markov blanket of the node"""
        if hasattr(self, 'markov_blanket'):
            for k,v in kwargs.items():
                if k in self.markov_blanket.keys():
                    print("Error: " + str(k) + " is already in the markov blanket of " + str(self))
                    exit(1)
                else:
                    self.markov_blanket[k] = v
        else:
            self.markov_blanket = kwargs

    def getMarkovBlanket(self):
        """ Method to return the Markov blanket of the node """
        return self.markov_blanket

    def update(self, ix=None, ro=1.):
        """ General method to update both parameters and expectations of the node """
        self.updateParameters(ix, ro)
        self.updateExpectations()

    def updateExpectations(self):
        """ General method to update the expectations of a node """
        pass

    def getDimensions(self):
        """ Method to return dimensionality of the node """
        return self.dim

    def getExpectations(self):
        """ General method to get all expectations of a node """
        pass

    def getExpectation(self):
        """ General method to get the first moment (expectation) of a node """
        pass

    def updateParameters(self, ix=None, ro=1.):
        """ General function to update parameters of the node """
        pass

    def getParameters(self):
        """ General function to get the parameters of the node """
        pass

    def updateDim(self, axis, new_dim):
        """ Method to update the dimensionality of a node
        PARAMETERS
        ----------
        axis:
        new_dim:
        """
        dim = list(self.dim)
        dim[axis] = new_dim
        self.dim = tuple(dim)

    def precompute(self, options=None):
        pass


class Constant_Node(Node):
    """ General class for a constant node in a Bayesian network
    Constant nodes do not have expectations or parameters but just values.
    However, for technical reasons the constant cvalues are defined as expectations
    """
    def __init__(self, dim, value):
        self.dim = dim
        if isinstance(value,(int,float)):
            if config.use_float32:
                self.value = value * np.ones(dim, dtype=np.float32)
            else:
                self.value = value * np.ones(dim, dtype=np.float64)
        else:
            assert value.shape == dim, "dimensionality mismatch"
            self.value = value

    def getValue(self):
        """ Method to return the values of the node """
        return self.value

    def getExpectation(self):
        """ Method to return the first moment of the node, which just points to the values """
        return self.getValue()

    def getExpectations(self):
        """ Method to return the expectations of the node, which just points to the values """
        return { 'E':self.getValue(), 'lnE':np.log(self.getValue()), 'E2':self.getValue()**2 }

    def removeFactors(self, idx, axis=None):
        if hasattr(self,"factors_axis"): axis = self.factors_axis
        if axis is not None:
            self.value = np.delete(self.value, idx, axis)
            self.updateDim(axis=axis, new_dim=self.dim[axis]-len(idx))

    def sample(self, distrib='P'):
        self.samp = self.value
        return self.samp



"""
The function cytotrace2 is a Python implementation of CytoTRACE 2, a deep learning-based tool for cell potency prediction. CytoTRACE 2 predicts
"""
import concurrent.futures
import math
import os
import warnings

import numpy as np
import pandas as pd
import scanpy as sc


from .._settings import add_reference

from ..externel.cytotrace2.gen_utils import *
#from cytotrace2_py.common.argument_parser import *

#pylint: disable=too-many-arguments
#pylint: disable=too-many-locals
def process_subset(idx, chunked_expression, smooth_batch_size, smooth_cores_to_use, 
                   species, use_model_dir, output_dir, max_pcs, seed):
    r"""Process a subset of the data in parallel for CytoTRACE2 prediction.
    
    Arguments:
        idx (int): Index of the current batch
        chunked_expression: Expression data chunk for this batch
        smooth_batch_size (int): Batch size for smoothing operations
        smooth_cores_to_use (int): Number of cores to use for smoothing
        species (str): Species type ('mouse' or 'human')
        use_model_dir (str): Path to model directory
        output_dir (str): Output directory for results
        max_pcs (int): Maximum number of principal components
        seed (int): Random seed for reproducibility
        
    Returns:
        pd.DataFrame: Smoothed CytoTRACE2 scores for the data chunk
    """

    # map and rank
    cell_names, gene_names, ranked_data = preprocess(chunked_expression, species)

    # top variable genes
    top_col_inds = top_var_genes(ranked_data)
    top_col_names = gene_names[top_col_inds]

    # predict by unrandomized chunked batches
    predicted_df = predict(ranked_data, cell_names, use_model_dir , chunked_expression.shape[0])

    smooth_score = smoothing_by_diffusion(predicted_df, ranked_data, 
                                          top_col_inds, smooth_batch_size,  seed)

    binned_score_pred_df = binning(predicted_df, smooth_score)

    # Transpose the matrix and create a DataFrame
    ranked_df = pd.DataFrame(ranked_data.T,  columns = cell_names)

    # Set the column names
    ranked_df.index = gene_names
    suffix = '_'+str(idx)
    binned_score_pred_df.to_csv(output_dir+'/binned_df'+suffix+'.txt',sep='\t', index=True)
    ranked_df.to_csv(output_dir+'/ranked_df'+suffix+'.txt', sep='\t', index=True)
	
    with open(output_dir+'/top_var_genes'+suffix+'.txt', 'w',
              encoding='utf-8') as f:
        for item in top_col_names:
            f.write(f"{item}\n")

    if chunked_expression.shape[0] < 100:
        print('cytotrace2: Fewer than 100 cells in dataset. Skipping KNN smoothing step.')
        smooth_by_knn_df = binned_score_pred_df.copy()
    else:
        #run_script = pkg_resources.resource_filename("cytotrace2_py","resources/smoothDatakNN.R")
        #knn_path = output_dir+'/smoothbykNNresult'+suffix+'.txt'
        #out = subprocess.run(['Rscript', run_script, '--output-dir', output_dir,
               #'--suffix', suffix, '--max-pcs', str(max_pcs), '--seed', str(seed)], check=True)
        #smooth_by_knn_df = pd.read_csv(knn_path, index_col = 0, sep='\t')
        from ..externel.cytotrace2.smoothDatakNN import smooth_data_kNN
        smooth_data_kNN(output_dir=output_dir, suffix=suffix, max_pcs=max_pcs, seed=seed)
        # 读取平滑处理后的结果
        knn_path = os.path.join(output_dir, f'smoothbykNNresult{suffix}.txt')
        smooth_by_knn_df = pd.read_csv(knn_path, index_col=0, sep='\t')

    return smooth_by_knn_df

def calculate_cores_to_use(chunk_number,smooth_chunk_number,max_cores,disable_parallelization):
    r"""Calculate optimal number of cores for parallel processing.
    
    Arguments:
        chunk_number (int): Number of data chunks
        smooth_chunk_number (int): Number of smoothing chunks
        max_cores (int): Maximum allowed cores
        disable_parallelization (bool): Whether to disable parallel processing
        
    Returns:
        tuple: (pred_cores_to_use, smooth_cores_to_use) number of cores for prediction and smoothing
    """

    pred_cores_to_use = 1
    smooth_cores_to_use = 1
    if smooth_chunk_number == 1 and chunk_number == 1:
        print("cytotrace2: The number of cells in your dataset is \
              less than the specified batch size.\n")
        print("    Model prediction will not be parallelized.")

    elif not disable_parallelization:
        # Calculate number of available processors
        num_proc = os.cpu_count()
        print("cytotrace2: "+str(num_proc)+" cores detected")
        if num_proc == 1:
            print("cytotrace2: Only one core detected. CytoTRACE 2 will not be run in parallel.")
        elif max_cores is None:
            pred_cores_to_use = min(chunk_number,num_proc-1)
            smooth_cores_to_use = min(smooth_chunk_number,max(math.floor((num_proc-1)/pred_cores_to_use),1))
            print('cytotrace2: Running '+str(pred_cores_to_use)+' prediction batch(es) in parallel using '+str(smooth_cores_to_use)+' cores for smoothing per batch.')
        else:
            max_cores = min(max_cores,num_proc-1)
            pred_cores_to_use = min(chunk_number,max_cores)
            smooth_cores_to_use = min(smooth_chunk_number,
                                      max(math.floor(max_cores/pred_cores_to_use),1))
            print('cytotrace2: Running '+str(pred_cores_to_use)+' prediction batch(es) in parallel using '+str(smooth_cores_to_use)+' cores for smoothing per batch.')

    return pred_cores_to_use, smooth_cores_to_use


def cytotrace2(
               adata,
               use_model_dir,
               species = "mouse",
               #full_model = False,
               batch_size = 10000,
               smooth_batch_size = 1000,
               disable_parallelization = False,
               max_cores = None,
               max_pcs = 200,
               seed = 14,
               output_dir = 'cytotrace2_results'):
    r"""CytoTRACE 2: Deep learning-based cell potency prediction.

    This function predicts cellular potency states using a deep learning model,
    providing scores for developmental potential and differentiation status.

    Arguments:
        adata: AnnData object containing single-cell RNA-seq data
        use_model_dir (str): Path to directory containing pre-trained model files
        species (str): Species of input data - 'mouse' or 'human' (default: 'mouse')
        batch_size (int): Number of cells to process per batch (default: 10000)
        smooth_batch_size (int): Batch size for smoothing operations (default: 1000)
        disable_parallelization (bool): Whether to disable parallel processing (default: False)
        max_cores (int): Maximum CPU cores for parallel processing (default: None for all cores)
        max_pcs (int): Maximum number of principal components to use (default: 200)
        seed (int): Random seed for reproducibility (default: 14)
        output_dir (str): Directory to save results (default: 'cytotrace2_results')

    Returns:
        pd.DataFrame: DataFrame containing CytoTRACE2 scores, potency categories, and relative scores
    
    """

    # Make output directory 
    out = os.system('mkdir -p '+output_dir)

    # Load data
    print('cytotrace2: Input parameters')
    #print('    Input file: '+input_path)
    print('    Species: '+species)
    #print('    Full model: '+str(full_model))
    print('    Parallelization enabled: '+str(not disable_parallelization))
    print('    User-provided limit for number of cores to use: '+str(max_cores))
    print('    Batch size: '+str(batch_size))
    print('    Smoothing batch size: '+str(smooth_batch_size))
    print('    Max PCs: '+str(max_pcs))
    print('    Seed: '+str(seed))
    print('    Output directory: '+output_dir)


    #expression =  adata.X.toarray()
    #if sparse.issparse(adata.X):
    from scipy.sparse import issparse
    if issparse(adata.X):
        expression=pd.DataFrame(adata.X.toarray(),index=adata.obs.index,columns=adata.var.index)
    else:
        expression = pd.DataFrame(adata.X, index=adata.obs.index, columns=adata.var.index)
    print('cytotrace2: Dataset characteristics')
    print('    Number of input genes: ',str(expression.shape[1]))
    print('    Number of input cells: ',str(expression.shape[0]))
    
    # Check if the input species is accurate
    # Calculate the proportion of row names that are all uppercase (assumed to be human) or not all uppercase (assumed to be mouse)
    is_human = sum([name.isupper() for name in expression.columns]) / expression.shape[1] > 0.9
    is_mouse = sum([not name.isupper() for name in expression.columns]) / expression.shape[1] > 0.9

    if is_human and species == 'mouse':
        warnings.warn("Species is most likely human. Please revise the 'species' input to the function.")

    if is_mouse and species == 'human':
        warnings.warn("Species is most likely mouse. Please revise the 'species' input to the function.")

    np.random.seed(seed)
    if batch_size > len(expression):
        print("cytotrace2: The passed batch_size is greater than the number of cells in the subsample. \n    Now setting batch_size to "+str(len(expression))+".")
        batch_size <- len(expression)
    if batch_size > 10000:
        print(".   Please consider reducing the batch size to 10000 for runtime and memory efficiency.")
    elif len(expression) > 10000 and batch_size > 10000:
        print("cytotrace2: Please consider reducing the batch_size to 10000 for runtime and memory efficiency.")
    
    print('cytotrace2: Preprocessing')
    
    # Calculate chunk number
    chunk_number = math.ceil(len(expression) / batch_size)
    smooth_chunk_number = math.ceil(batch_size/smooth_batch_size)
    if len(expression) < 1000:
        chunk_number = 1
        smooth_chunk_number = 1

    # Determine multiprocessing parameters
    pred_cores_to_use, smooth_cores_to_use = calculate_cores_to_use(chunk_number, smooth_chunk_number, max_cores, disable_parallelization)

    #if full_model == False:
    #    use_model_dir = pkg_resources.resource_filename("omicverse","data_files/5_models_weights/")
    #else:
    #    use_model_dir = pkg_resources.resource_filename("omicverse","data_files/17_models_weights/")
    
    #original_names = adata.var_names
    original_names = adata.obs_names
    subsamples_indices = np.arange(len(expression))
    np.random.shuffle(subsamples_indices)
    subsamples = np.array_split(subsamples_indices, chunk_number)
    
    predictions = []
   
    # Process each chunk separately
    results = []
    
    with concurrent.futures.ThreadPoolExecutor(max_workers=pred_cores_to_use) as executor:
        for idx in range(chunk_number):
            chunked_expression = expression.iloc[subsamples[idx], :]
            print('cytotrace2: Initiated processing batch ' + str(idx + 1) + '/' + str(chunk_number) + ' with ' + str(chunked_expression.shape[0]) + ' cells')
            results.append(executor.submit(process_subset, idx,
                                        chunked_expression, smooth_batch_size,
                                        smooth_cores_to_use, species,
                                        use_model_dir, output_dir, max_pcs, seed))
        
        for f in concurrent.futures.as_completed(results):
            smooth_by_knn_df = f.result()
            predictions.append(smooth_by_knn_df)
    
    

    for idx in range(chunk_number):
        suffix = '_'+str(idx)
        temp_file_list = [output_dir+'/binned_df'+suffix+'.txt',output_dir+'/ranked_df'+suffix+'.txt',
                                output_dir+'/smoothbykNNresult'+suffix+'.txt',output_dir+'/top_var_genes'+suffix+'.txt']
        for fin in temp_file_list:
            if os.path.isfile(fin):
                os.remove(fin)
    
    predicted_df_final = pd.concat(predictions, ignore_index=False)
    predicted_df_final = predicted_df_final.loc[original_names]
    ranges = np.linspace(0, 1, 7)  
    labels = [
        'Differentiated',
        'Unipotent',
        'Oligopotent',
        'Multipotent',
        'Pluripotent',
        'Totipotent']
    
    predicted_df_final['CytoTRACE2_Potency'] = pd.cut(predicted_df_final['CytoTRACE2_Score'], bins=ranges, labels=labels, include_lowest=True)

    ranked_scores = scipy.stats.rankdata(predicted_df_final['CytoTRACE2_Score'])
    relative_scores = (ranked_scores-min(ranked_scores))/(max(ranked_scores)-min(ranked_scores))
    predicted_df_final['CytoTRACE2_Relative'] = relative_scores

    adata.obs['CytoTRACE2_Score'] = predicted_df_final['CytoTRACE2_Score']
    adata.obs['CytoTRACE2_Potency'] = predicted_df_final['CytoTRACE2_Potency']
    adata.obs['CytoTRACE2_Relative'] = predicted_df_final['CytoTRACE2_Relative']
    adata.obs['preKNN_CytoTRACE2_Score']=predicted_df_final['preKNN_CytoTRACE2_Score']
    adata.obs['preKNN_CytoTRACE2_Potency']=predicted_df_final['preKNN_CytoTRACE2_Potency']
    print('cytotrace2: Results saved to adata.obs \
          \n    CytoTRACE2_Score: CytoTRACE2 score \
          \n    CytoTRACE2_Potency: CytoTRACE2 potency \
          \n    CytoTRACE2_Relative: CytoTRACE2 relative score \
          \n    preKNN_CytoTRACE2_Score: CytoTRACE2 score before kNN smoothing \
          \n    preKNN_CytoTRACE2_Potency: CytoTRACE2 potency before kNN smoothing')
    #predicted_df_final = predicted_df_final[["CytoTRACE2_Score", "CytoTRACE2_Potency" , "CytoTRACE2_Relative", "preKNN_CytoTRACE2_Score", "preKNN_CytoTRACE2_Potency"]]

    predicted_df_final.to_csv(output_dir+'/cytotrace2_results.txt',sep='\t')

    #print('cytotrace2: Plotting outputs')

    #plot_script = pkg_resources.resource_filename("cytotrace2_py","resources/plot_cytotrace2_results.R")
    
    #if annotation_path != "":
    #    out = subprocess.run(['Rscript', plot_script, '--expression-path', input_path, '--result-path', output_dir+'/cytotrace2_results.txt', '--annotation-path', annotation_path, '--plot-dir', output_dir+'/plots'], check=True)
    #else:
    #    out = subprocess.run(['Rscript', plot_script, '--expression-path', input_path, '--result-path', output_dir+'/cytotrace2_results.txt', '--plot-dir', output_dir+'/plots'], capture_output=True)

    print('cytotrace2: Finished.')
    add_reference(adata,'CytoTRACE2','cell potency prediction with CytoTRACE2')
    return predicted_df_final
